#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for database migrations for the API database.

These are "opportunistic" tests which allow testing against all three databases
(sqlite in memory, mysql, pg) in a properly configured unit test environment.

For the opportunistic testing you need to set up DBs named 'openstack_citest'
with user 'openstack_citest' and password 'openstack_citest' on localhost. The
test will then use that DB and username/password combo to run the tests. Refer
to the 'tools/test-setup.sh' for an example of how to configure this.
"""

from unittest import mock

from alembic import command as alembic_api
from alembic import script as alembic_script
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import test_fixtures
from oslo_db.sqlalchemy import test_migrations
from oslo_log import log as logging
import sqlalchemy
import testtools

from nova.db.api import models
from nova.db import migration
from nova import test

LOG = logging.getLogger(__name__)


class NovaModelsMigrationsSync(test_migrations.ModelsMigrationsSync):
    """Test that the models match the database after migrations are run."""

    # Migrations can take a long time, particularly on underpowered CI nodes.
    # Give them some breathing room.
    TIMEOUT_SCALING_FACTOR = 4

    def setUp(self):
        super().setUp()
        self.engine = enginefacade.writer.get_engine()

    def db_sync(self, engine):
        with mock.patch.object(migration, '_get_engine', return_value=engine):
            migration.db_sync(database='api')

    def get_engine(self):
        return self.engine

    def get_metadata(self):
        return models.BASE.metadata

    def include_object(self, object_, name, type_, reflected, compare_to):
        if type_ == 'table':
            # Define a whitelist of tables that will be removed from the DB in
            # a later release and don't have a corresponding model anymore.

            return name not in models.REMOVED_TABLES

        return True

    def filter_metadata_diff(self, diff):
        # Filter out diffs that shouldn't cause a sync failure.
        new_diff = []

        for element in diff:
            if isinstance(element, list):
                # modify_nullable is a list
                new_diff.append(element)
            else:
                # tuple with action as first element. Different actions have
                # different tuple structures.
                if element[0] == 'add_fk':
                    fkey = element[1]
                    tablename = fkey.table.name
                    column_keys = fkey.column_keys
                    if (tablename, column_keys) in models.REMOVED_FKEYS:
                        continue
                elif element[0] == 'remove_column':
                    table = element[2]
                    column = element[3].name
                    if (table, column) in models.REMOVED_COLUMNS:
                        continue

                new_diff.append(element)

        return new_diff


class TestModelsSyncSQLite(
    NovaModelsMigrationsSync,
    test_fixtures.OpportunisticDBTestMixin,
    testtools.TestCase,
):
    pass


class TestModelsSyncMySQL(
    NovaModelsMigrationsSync,
    test_fixtures.OpportunisticDBTestMixin,
    testtools.TestCase,
):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture


class TestModelsSyncPostgreSQL(
    NovaModelsMigrationsSync,
    test_fixtures.OpportunisticDBTestMixin,
    testtools.TestCase,
):
    FIXTURE = test_fixtures.PostgresqlOpportunisticFixture


class NovaMigrationsWalk(
    test_fixtures.OpportunisticDBTestMixin, test.NoDBTestCase,
):

    # Migrations can take a long time, particularly on underpowered CI nodes.
    # Give them some breathing room.
    TIMEOUT_SCALING_FACTOR = 4

    def setUp(self):
        super().setUp()
        self.engine = enginefacade.writer.get_engine()
        self.config = migration._find_alembic_conf('api')
        self.init_version = 'd67eeaabee36'

    def _migrate_up(self, connection, revision):
        if revision == self.init_version:  # no tests for the initial revision
            alembic_api.upgrade(self.config, revision)
            return

        self.assertIsNotNone(
            getattr(self, '_check_%s' % revision, None),
            (
                'API DB Migration %s does not have a test; you must add one'
            ) % revision,
        )

        pre_upgrade = getattr(self, '_pre_upgrade_%s' % revision, None)
        if pre_upgrade:
            pre_upgrade(connection)

        alembic_api.upgrade(self.config, revision)

        post_upgrade = getattr(self, '_check_%s' % revision, None)
        if post_upgrade:
            post_upgrade(connection)

    _b30f573d3377_removed_columns = {
        'access_ip_v4',
        'access_ip_v6',
        'config_drive',
        'display_name',
        'image_ref',
        'info_cache',
        'instance_metadata',
        'key_name',
        'locked_by',
        'progress',
        'request_spec_id',
        'security_groups',
        'task_state',
        'user_id',
        'vm_state',
    }

    def _pre_upgrade_b30f573d3377(self, connection):
        # we use the inspector here rather than oslo_db.utils.column_exists,
        # since the latter will create a new connection
        inspector = sqlalchemy.inspect(connection)
        columns = [x['name'] for x in inspector.get_columns('build_requests')]
        for removed_column in self._b30f573d3377_removed_columns:
            self.assertIn(removed_column, columns)

    def _check_b30f573d3377(self, connection):
        # we use the inspector here rather than oslo_db.utils.column_exists,
        # since the latter will create a new connection
        inspector = sqlalchemy.inspect(connection)
        columns = [x['name'] for x in inspector.get_columns('build_requests')]
        for removed_column in self._b30f573d3377_removed_columns:
            self.assertNotIn(removed_column, columns)

    def _check_cdeec0c85668(self, connection):
        # the table optionally existed: there's no way to check for its
        # removal without creating it first, which is dumb
        pass

    def test_single_base_revision(self):
        """Ensure we only have a single base revision.

        There's no good reason for us to have diverging history, so validate
        that only one base revision exists. This will prevent simple errors
        where people forget to specify the base revision. If this fail for your
        change, look for migrations that do not have a 'revises' line in them.
        """
        script = alembic_script.ScriptDirectory.from_config(self.config)
        self.assertEqual(1, len(script.get_bases()))

    def test_single_head_revision(self):
        """Ensure we only have a single head revision.

        There's no good reason for us to have diverging history, so validate
        that only one head revision exists. This will prevent merge conflicts
        adding additional head revision points. If this fail for your change,
        look for migrations with the same 'revises' line in them.
        """
        script = alembic_script.ScriptDirectory.from_config(self.config)
        self.assertEqual(1, len(script.get_heads()))

    def test_walk_versions(self):
        with self.engine.begin() as connection:
            self.config.attributes['connection'] = connection
            script = alembic_script.ScriptDirectory.from_config(self.config)
            revisions = [x.revision for x in script.walk_revisions()]

            # for some reason, 'walk_revisions' gives us the revisions in
            # reverse chronological order so we have to invert this
            revisions.reverse()
            self.assertEqual(revisions[0], self.init_version)

            for revision in revisions:
                LOG.info('Testing revision %s', revision)
                self._migrate_up(connection, revision)

    def test_db_version_alembic(self):
        engine = enginefacade.writer.get_engine()

        with mock.patch.object(migration, '_get_engine', return_value=engine):
            migration.db_sync(database='api')

            script = alembic_script.ScriptDirectory.from_config(self.config)
            head = script.get_current_head()
            self.assertEqual(head, migration.db_version(database='api'))


class TestMigrationsWalkSQLite(
    NovaMigrationsWalk,
    test_fixtures.OpportunisticDBTestMixin,
    test.NoDBTestCase,
):
    pass


class TestMigrationsWalkMySQL(
    NovaMigrationsWalk,
    test_fixtures.OpportunisticDBTestMixin,
    test.NoDBTestCase,
):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture


class TestMigrationsWalkPostgreSQL(
    NovaMigrationsWalk,
    test_fixtures.OpportunisticDBTestMixin,
    test.NoDBTestCase,
):
    FIXTURE = test_fixtures.PostgresqlOpportunisticFixture



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import webob

from nova.api.openstack.compute import server_shares
from nova.compute import vm_states
from nova import context
from nova.db.main import models
from nova import objects
from nova.tests.unit.api.openstack import fakes
from nova.tests.unit.compute.test_compute import BaseTestCase
from nova.tests.unit import fake_instance

from nova.tests import fixtures as nova_fixtures
from oslo_utils import timeutils

from unittest import mock

UUID = 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'
NON_EXISTING_UUID = '123'


def return_server(compute_api, context, instance_id, expected_attrs=None):
    return fake_instance.fake_instance_obj(context, vm_state=vm_states.ACTIVE)


def return_invalid_server(compute_api, context, instance_id,
                          expected_attrs=None):
    return fake_instance.fake_instance_obj(context,
                                           vm_state=vm_states.BUILDING)


class ServerSharesTest(BaseTestCase):
    wsgi_api_version = '2.97'

    def setUp(self):
        super(ServerSharesTest, self).setUp()
        self.controller = server_shares.ServerSharesController()
        inst_map = objects.InstanceMapping(
            project_id=fakes.FAKE_PROJECT_ID,
            user_id=fakes.FAKE_USER_ID,
            cell_mapping=objects.CellMappingList.get_all(
                context.get_admin_context())[1])
        self.stub_out('nova.objects.InstanceMapping.get_by_instance_uuid',
                      lambda s, c, u: inst_map)
        self.req = fakes.HTTPRequest.blank(
                '/servers/%s/shares' % (UUID),
                use_admin_context=False, version=self.wsgi_api_version)
        self.manila_fixture = self.useFixture(nova_fixtures.ManilaFixture())

    def fake_get_instance(self):
        ctxt = self.req.environ['nova.context']
        return fake_instance.fake_instance_obj(
                ctxt,
                uuid=fakes.FAKE_UUID,
                flavor = objects.Flavor(id=1, name='flavor1',
                    memory_mb=256, vcpus=1,
                    root_gb=1, ephemeral_gb=1,
                    flavorid='1',
                    swap=0, rxtx_factor=1.0,
                    vcpu_weight=1,
                    disabled=False,
                    is_public=True,
                    extra_specs={
                        'virtiofs': 'required',
                        'mem_backing_file': 'required'
                        },
                    projects=[]),
                vm_state=vm_states.STOPPED)

    @mock.patch(
        'nova.virt.hardware.check_shares_supported', return_value=None
    )
    @mock.patch('nova.db.main.api.share_mapping_get_by_instance_uuid')
    @mock.patch('nova.api.openstack.common.get_instance')
    def test_index(
            self, mock_get_instance, mock_db_get_shares, mock_shares_support
    ):
        timeutils.set_time_override()
        NOW = timeutils.utcnow()
        instance = self.fake_get_instance()
        mock_get_instance.return_value = instance

        fake_db_shares = [
            {
                'created_at': NOW,
                'updated_at': None,
                'deleted_at': None,
                'deleted': False,
                "id": 1,
                "uuid": "33a8e0cb-5f82-409a-b310-89c41f8bf023",
                "instance_uuid": "48c16a1a-183f-4052-9dac-0e4fc1e498ae",
                "share_id": "48c16a1a-183f-4052-9dac-0e4fc1e498ad",
                "status": "active",
                "tag": "foo",
                "export_location": "10.0.0.50:/mnt/foo",
                "share_proto": "NFS",
            },
            {
                'created_at': NOW,
                'updated_at': None,
                'deleted_at': None,
                'deleted': False,
                "id": 2,
                "uuid": "33a8e0cb-5f82-409a-b310-89c41f8bf024",
                "instance_uuid": "48c16a1a-183f-4052-9dac-0e4fc1e498ae",
                "share_id": "e8debdc0-447a-4376-a10a-4cd9122d7986",
                "status": "active",
                "tag": "bar",
                "export_location": "10.0.0.50:/mnt/bar",
                "share_proto": "NFS",
            }
        ]

        fake_shares = {
            "shares": [
                {
                    "share_id": "48c16a1a-183f-4052-9dac-0e4fc1e498ad",
                    "status": "active",
                    "tag": "foo",
                },
                {
                    "share_id": "e8debdc0-447a-4376-a10a-4cd9122d7986",
                    "status": "active",
                    "tag": "bar",
                }
            ]
        }

        mock_db_get_shares.return_value = fake_db_shares
        output = self.controller.index(self.req, instance.uuid)
        mock_db_get_shares.assert_called_once_with(mock.ANY, instance.uuid)
        self.assertEqual(output, fake_shares)

    @mock.patch(
        'nova.virt.hardware.check_shares_supported', return_value=None
    )
    @mock.patch(
        'nova.compute.utils.notify_about_share_attach_detach',
        return_value=None
    )
    @mock.patch(
        'nova.db.main.api.share_mapping_get_by_instance_uuid_and_share_id'
    )
    @mock.patch('nova.db.main.api.share_mapping_update')
    @mock.patch('nova.api.openstack.common.get_instance')
    def test_create(
        self,
        mock_get_instance,
        mock_db_update_share,
        mock_db_get_share,
        mock_notifications,
        mock_shares_support,
    ):
        instance = self.fake_get_instance()

        mock_get_instance.return_value = instance

        fake_db_share = {
            'created_at': None,
            'updated_at': None,
            'deleted_at': None,
            'deleted': False,
            "id": 1,
            "uuid": "7ddcf3ae-82d4-4f93-996a-2b6cbcb42c2b",
            "instance_uuid": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "share_id": "e8debdc0-447a-4376-a10a-4cd9122d7986",
            "status": "attaching",
            "tag": "e8debdc0-447a-4376-a10a-4cd9122d7986",
            "export_location": "10.0.0.50:/mnt/foo",
            "share_proto": "NFS",
        }

        body = {
            'share': {
                'share_id': 'e8debdc0-447a-4376-a10a-4cd9122d7986'
            }}

        mock_db_update_share.return_value = fake_db_share
        mock_db_get_share.side_effect = [None, fake_db_share]
        self.controller.create(self.req, instance.uuid, body=body)

        mock_db_update_share.assert_called_once_with(
            mock.ANY,
            mock.ANY,
            instance.uuid,
            fake_db_share['share_id'],
            'attaching',
            fake_db_share['tag'],
            fake_db_share['export_location'],
            fake_db_share['share_proto'],
        )

    @mock.patch('nova.compute.api.API.allow_share')
    @mock.patch(
        'nova.virt.hardware.check_shares_supported', return_value=None
    )
    @mock.patch(
        'nova.db.main.api.share_mapping_get_by_instance_uuid_and_share_id'
    )
    @mock.patch('nova.db.main.api.share_mapping_update')
    @mock.patch('nova.api.openstack.common.get_instance')
    def test_create_share_with_new_tag(
        self,
        mock_get_instance,
        mock_db_update_share,
        mock_db_get_share,
        mock_shares_support,
        mock_allow
    ):
        instance = self.fake_get_instance()

        mock_get_instance.return_value = instance

        fake_db_share = {
            'created_at': None,
            'updated_at': None,
            'deleted_at': None,
            'deleted': False,
            "id": 1,
            "uuid": "7ddcf3ae-82d4-4f93-996a-2b6cbcb42c2b",
            "instance_uuid": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "share_id": "e8debdc0-447a-4376-a10a-4cd9122d7986",
            "status": "attaching",
            "tag": "e8debdc0-447a-4376-a10a-4cd9122d7986",
            "export_location": "10.0.0.50:/mnt/foo",
            "share_proto": "NFS",
        }

        body = {
            'share': {
                'share_id': 'e8debdc0-447a-4376-a10a-4cd9122d7986'
            }}

        mock_db_update_share.return_value = fake_db_share
        mock_db_get_share.side_effect = [None, fake_db_share]
        self.controller.create(self.req, instance.uuid, body=body)

        mock_allow.assert_called_once()
        self.assertIsInstance(
            mock_allow.call_args.args[1], objects.instance.Instance)
        self.assertEqual(mock_allow.call_args.args[1].uuid, instance.uuid)
        self.assertIsInstance(
            mock_allow.call_args.args[2], objects.share_mapping.ShareMapping)
        self.assertEqual(
            mock_allow.call_args.args[2].share_id, fake_db_share['share_id'])

        mock_db_update_share.assert_called_once_with(
            mock.ANY,
            mock.ANY,
            instance.uuid,
            fake_db_share['share_id'],
            'attaching',
            fake_db_share['tag'],
            fake_db_share['export_location'],
            fake_db_share['share_proto'],
        )

        # Change the tag of the share
        body['share']['tag'] = 'my-tag'
        mock_db_update_share.return_value['tag'] = "my-tag"
        mock_db_get_share.side_effect = [
            fake_db_share,
            mock_db_update_share.return_value,
        ]

        exc = self.assertRaises(
            webob.exc.HTTPConflict,
            self.controller.create,
            self.req,
            instance.uuid,
            body=body,
        )

        self.assertIn(
            "Share 'e8debdc0-447a-4376-a10a-4cd9122d7986' or tag 'my-tag' "
            "already associated to this server",
            str(exc))

    @mock.patch(
        'nova.virt.hardware.check_shares_supported', return_value=None
    )
    @mock.patch('nova.api.openstack.common.get_instance')
    def test_create_passing_a_share_with_an_error(
        self,
        mock_get_instance,
        mock_shares_support,
    ):
        instance = self.fake_get_instance()

        mock_get_instance.return_value = instance

        body = {
            'share': {
                'share_id': 'e8debdc0-447a-4376-a10a-4cd9122d7986'
            }}

        self.manila_fixture.mock_get.side_effect = (
            self.manila_fixture.fake_get_share_status_error
        )

        exc = self.assertRaises(
            webob.exc.HTTPConflict,
            self.controller.create,
            self.req,
            instance.uuid,
            body=body,
        )
        self.assertEqual(
            str(exc),
            "Share e8debdc0-447a-4376-a10a-4cd9122d7986 is in 'error' "
            "instead of 'available' status.",
        )

    @mock.patch(
        'nova.virt.hardware.check_shares_supported', return_value=None
    )
    @mock.patch('nova.api.openstack.common.get_instance')
    def test_create_passing_unknown_protocol(
        self,
        mock_get_instance,
        mock_shares_support,
    ):
        instance = self.fake_get_instance()

        mock_get_instance.return_value = instance

        body = {
            'share': {
                'share_id': 'e8debdc0-447a-4376-a10a-4cd9122d7986'
            }}

        self.manila_fixture.mock_get.side_effect = (
            self.manila_fixture.fake_get_share_unknown_protocol
        )

        exc = self.assertRaises(
            webob.exc.HTTPConflict,
            self.controller.create,
            self.req,
            instance.uuid,
            body=body,
        )
        self.assertEqual(
            str(exc),
            "Share protocol CIFS is not supported."
        )

    @mock.patch('nova.compute.api.API.deny_share')
    @mock.patch(
        'nova.virt.hardware.check_shares_supported', return_value=None
    )
    @mock.patch(
        'nova.compute.utils.notify_about_share_attach_detach',
        return_value=None
    )
    @mock.patch('nova.db.main.api.'
            'share_mapping_delete_by_instance_uuid_and_share_id')
    @mock.patch('nova.db.main.api.'
            'share_mapping_get_by_instance_uuid_and_share_id')
    @mock.patch('nova.api.openstack.common.get_instance')
    def test_delete(
        self,
        mock_get_instance,
        mock_db_get_shares,
        mock_db_delete_share,
        mock_notifications,
        mock_shares_support,
        mock_deny
    ):
        instance = self.fake_get_instance()

        mock_get_instance.return_value = instance

        fake_db_share = models.ShareMapping()
        fake_db_share.created_at = None
        fake_db_share.updated_at = None
        fake_db_share.deleted_at = None
        fake_db_share.deleted = False
        fake_db_share.id = 1
        fake_db_share.uuid = "33a8e0cb-5f82-409a-b310-89c41f8bf023"
        fake_db_share.instance_uuid = "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"
        fake_db_share.share_id = "e8debdc0-447a-4376-a10a-4cd9122d7986"
        fake_db_share.status = "inactive"
        fake_db_share.tag = "e8debdc0-447a-4376-a10a-4cd9122d7986"
        fake_db_share.export_location = "10.0.0.50:/mnt/foo"
        fake_db_share.share_proto = "NFS"

        mock_db_get_shares.return_value = fake_db_share
        self.controller.delete(
                self.req, instance.uuid, fake_db_share.share_id)

        mock_deny.assert_called_once()
        self.assertIsInstance(
            mock_deny.call_args.args[1], objects.instance.Instance)
        self.assertEqual(mock_deny.call_args.args[1].uuid, instance.uuid)
        self.assertIsInstance(
            mock_deny.call_args.args[2], objects.share_mapping.ShareMapping)
        self.assertEqual(
            mock_deny.call_args.args[2].share_id, fake_db_share['share_id'])



# Copyright 2016 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
CLI interface for nova status commands.
"""

import functools
import sys
import traceback

from keystoneauth1 import exceptions as ks_exc
import microversion_parse
from oslo_config import cfg
from oslo_upgradecheck import common_checks
from oslo_upgradecheck import upgradecheck
import sqlalchemy as sa
from sqlalchemy import func as sqlfunc

from nova.cmd import common as cmd_common
import nova.conf
from nova import config
from nova import context as nova_context
from nova.db.api import api as api_db_api
from nova.db.main import api as main_db_api
from nova import exception
from nova.i18n import _
from nova.objects import cell_mapping as cell_mapping_obj
# NOTE(lyarwood): The following are imported as machine_type_utils expects them
# to be registered under nova.objects when called via _check_machine_type_set
from nova.objects import image_meta as image_meta_obj  # noqa: F401
from nova.objects import instance as instance_obj  # noqa: F401
from nova import utils
from nova import version
from nova.virt.libvirt import machine_type_utils
from nova.volume import cinder

CONF = nova.conf.CONF

# NOTE(gibi): 1.36 is required by nova-scheduler to support the same_subtree
# queryparam to make GET /allocation_candidates require that a list of request
# groups are satisfied from the same provider subtree.
# NOTE: If you bump this version, remember to update the history
# section in the nova-status man page (doc/source/cli/nova-status).
MIN_PLACEMENT_MICROVERSION = "1.36"

# NOTE(mriedem): 3.44 is needed to work with volume attachment records which
# are required for supporting multi-attach capable volumes.
MIN_CINDER_MICROVERSION = '3.44'


class UpgradeCommands(upgradecheck.UpgradeCommands):
    """Commands related to upgrades.

    The subcommands here must not rely on the nova object model since they
    should be able to run on n-1 data. Any queries to the database should be
    done through the sqlalchemy query language directly like the database
    schema migrations.
    """

    def _count_compute_nodes(self, context=None):
        """Returns the number of compute nodes in the cell database."""
        # NOTE(mriedem): This does not filter based on the service status
        # because a disabled nova-compute service could still be reporting
        # inventory info to the placement service. There could be an outside
        # chance that there are compute node records in the database for
        # disabled nova-compute services that aren't yet upgraded to Ocata or
        # the nova-compute service was deleted and the service isn't actually
        # running on the compute host but the operator hasn't cleaned up the
        # compute_nodes entry in the database yet. We consider those edge cases
        # here and the worst case scenario is we give a warning that there are
        # more compute nodes than resource providers. We can tighten this up
        # later if needed, for example by not including compute nodes that
        # don't have a corresponding nova-compute service in the services
        # table, or by only counting compute nodes with a service version of at
        # least 15 which was the highest service version when Newton was
        # released.
        meta = sa.MetaData()
        engine = main_db_api.get_engine(context=context)
        compute_nodes = sa.Table('compute_nodes', meta, autoload_with=engine)
        with engine.connect() as conn:
            return conn.execute(
                sa.select(sqlfunc.count()).select_from(compute_nodes).where(
                    compute_nodes.c.deleted == 0
                )
            ).scalars().first()

    def _check_cellsv2(self):
        """Checks to see if cells v2 has been setup.

        These are the same checks performed in the 030_require_cell_setup API
        DB migration except we expect this to be run AFTER the
        nova-manage cell_v2 simple_cell_setup command, which would create the
        cell and host mappings and sync the cell0 database schema, so we don't
        check for flavors at all because you could create those after doing
        this on an initial install. This also has to be careful about checking
        for compute nodes if there are no host mappings on a fresh install.
        """
        meta = sa.MetaData()
        engine = api_db_api.get_engine()

        cell_mappings = self._get_cell_mappings()
        count = len(cell_mappings)
        # Two mappings are required at a minimum, cell0 and your first cell
        if count < 2:
            msg = _('There needs to be at least two cell mappings, one for '
                    'cell0 and one for your first cell. Run command '
                    '\'nova-manage cell_v2 simple_cell_setup\' and then '
                    'retry.')
            return upgradecheck.Result(upgradecheck.Code.FAILURE, msg)

        cell0 = any(mapping.is_cell0() for mapping in cell_mappings)
        if not cell0:
            msg = _('No cell0 mapping found. Run command '
                    '\'nova-manage cell_v2 simple_cell_setup\' and then '
                    'retry.')
            return upgradecheck.Result(upgradecheck.Code.FAILURE, msg)

        host_mappings = sa.Table('host_mappings', meta, autoload_with=engine)

        with engine.connect() as conn:
            count = conn.execute(
                sa.select(sqlfunc.count()).select_from(host_mappings)
            ).scalars().first()

        if count == 0:
            # This may be a fresh install in which case there may not be any
            # compute_nodes in the cell database if the nova-compute service
            # hasn't started yet to create those records. So let's query the
            # cell database for compute_nodes records and if we find at least
            # one it's a failure.
            num_computes = self._count_compute_nodes()
            if num_computes > 0:
                msg = _('No host mappings found but there are compute nodes. '
                        'Run command \'nova-manage cell_v2 '
                        'simple_cell_setup\' and then retry.')
                return upgradecheck.Result(upgradecheck.Code.FAILURE, msg)

            msg = _('No host mappings or compute nodes were found. Remember '
                    'to run command \'nova-manage cell_v2 discover_hosts\' '
                    'when new compute hosts are deployed.')
            return upgradecheck.Result(upgradecheck.Code.SUCCESS, msg)

        return upgradecheck.Result(upgradecheck.Code.SUCCESS)

    @staticmethod
    def _placement_get(path):
        """Do an HTTP get call against placement engine.

        This is in a dedicated method to make it easier for unit
        testing purposes.

        """
        client = utils.get_ksa_adapter('placement')
        return client.get(path, raise_exc=True).json()

    def _check_placement(self):
        """Checks to see if the placement API is ready for scheduling.

        Checks to see that the placement API service is registered in the
        service catalog and that we can make requests against it.
        """
        try:
            # TODO(efried): Use ksa's version filtering in _placement_get
            versions = self._placement_get("/")
            max_version = microversion_parse.parse_version_string(
                versions["versions"][0]["max_version"])
            needs_version = microversion_parse.parse_version_string(
                MIN_PLACEMENT_MICROVERSION)
            if max_version < needs_version:
                msg = (_('Placement API version %(needed)s needed, '
                         'you have %(current)s.') %
                       {'needed': needs_version, 'current': max_version})
                return upgradecheck.Result(upgradecheck.Code.FAILURE, msg)
        except ks_exc.MissingAuthPlugin:
            msg = _('No credentials specified for placement API in nova.conf.')
            return upgradecheck.Result(upgradecheck.Code.FAILURE, msg)
        except ks_exc.Unauthorized:
            msg = _('Placement service credentials do not work.')
            return upgradecheck.Result(upgradecheck.Code.FAILURE, msg)
        except ks_exc.EndpointNotFound:
            msg = _('Placement API endpoint not found.')
            return upgradecheck.Result(upgradecheck.Code.FAILURE, msg)
        except ks_exc.DiscoveryFailure:
            msg = _('Discovery for placement API URI failed.')
            return upgradecheck.Result(upgradecheck.Code.FAILURE, msg)
        except ks_exc.NotFound:
            msg = _('Placement API does not seem to be running.')
            return upgradecheck.Result(upgradecheck.Code.FAILURE, msg)

        return upgradecheck.Result(upgradecheck.Code.SUCCESS)

    @staticmethod
    def _get_cell_mappings():
        """Queries the API database for cell mappings.

        .. note:: This method is unique in that it queries the database using
                  CellMappingList.get_all() rather than a direct query using
                  the sqlalchemy models. This is because
                  CellMapping.database_connection can be a template and the
                  object takes care of formatting the URL. We cannot use
                  RowProxy objects from sqlalchemy because we cannot set the
                  formatted 'database_connection' value back on those objects
                  (they are read-only).

        :returns: list of nova.objects.CellMapping objects
        """
        ctxt = nova_context.get_admin_context()
        cell_mappings = cell_mapping_obj.CellMappingList.get_all(ctxt)
        return cell_mappings

    def _check_cinder(self):
        """Checks to see that the cinder API is available at a given minimum
        microversion.
        """
        # Check to see if nova is even configured for Cinder yet (fresh install
        # or maybe not using Cinder at all).
        if CONF.cinder.auth_type is None:
            return upgradecheck.Result(upgradecheck.Code.SUCCESS)

        try:
            # TODO(mriedem): Eventually use get_ksa_adapter here when it
            # supports cinder.
            cinder.is_microversion_supported(
                nova_context.get_admin_context(), MIN_CINDER_MICROVERSION)
        except exception.CinderAPIVersionNotAvailable:
            return upgradecheck.Result(
                upgradecheck.Code.FAILURE,
                _('Cinder API %s or greater is required. Deploy at least '
                  'Cinder 12.0.0 (Queens).') % MIN_CINDER_MICROVERSION)
        except Exception as ex:
            # Anything else trying to connect, like bad config, is out of our
            # hands so just return a warning.
            return upgradecheck.Result(
                upgradecheck.Code.WARNING,
                _('Unable to determine Cinder API version due to error: %s') %
                str(ex))
        return upgradecheck.Result(upgradecheck.Code.SUCCESS)

    def _check_old_computes(self):
        # warn if there are computes in the system older than the previous
        # major release
        try:
            utils.raise_if_old_compute()
        except exception.TooOldComputeService as e:
            return upgradecheck.Result(upgradecheck.Code.FAILURE, str(e))

        return upgradecheck.Result(upgradecheck.Code.SUCCESS)

    def _check_machine_type_set(self):
        ctxt = nova_context.get_admin_context()
        if machine_type_utils.get_instances_without_type(ctxt):
            msg = (_("""
Instances found without hw_machine_type set. This warning can be ignored if
your environment does not contain libvirt based compute hosts.
Use the `nova-manage libvirt list_unset_machine_type` command to list these
instances. For more details see the following:
https://docs.openstack.org/nova/latest/admin/hw-machine-type.html"""))
            return upgradecheck.Result(upgradecheck.Code.WARNING, msg)

        return upgradecheck.Result(upgradecheck.Code.SUCCESS)

    def _check_service_user_token(self):
        if not CONF.service_user.send_service_user_token:
            msg = (_("""
Service user token configuration is required for all Nova services.
For more details see the following:
https://docs.openstack.org/nova/latest/admin/configuration/service-user-token.html"""))  # noqa
            return upgradecheck.Result(upgradecheck.Code.FAILURE, msg)
        return upgradecheck.Result(upgradecheck.Code.SUCCESS)

    # The format of the check functions is to return an upgradecheck.Result
    # object with the appropriate upgradecheck.Code and details set. If the
    # check hits warnings or failures then those should be stored in the
    # returned upgradecheck.Result's "details" attribute. The summary will
    # be rolled up at the end of the check() function. These functions are
    # intended to be run in order and build on top of each other so order
    # matters.
    _upgrade_checks = (
        # Added in Ocata
        (_('Cells v2'), _check_cellsv2),
        # Added in Ocata
        (_('Placement API'), _check_placement),
        # Added in Train
        (_('Cinder API'), _check_cinder),
        # Added in Victoria
        (
            _('Policy File JSON to YAML Migration'),
            (common_checks.check_policy_json, {'conf': CONF})
        ),
        # Added in Wallaby
        (_('Older than N-1 computes'), _check_old_computes),
        # Added in Wallaby
        (_('hw_machine_type unset'), _check_machine_type_set),
        # Added in Bobcat
        (_('Service User Token Configuration'), _check_service_user_token),
    )


CATEGORIES = {
    'upgrade': UpgradeCommands,
}


add_command_parsers = functools.partial(cmd_common.add_command_parsers,
                                        categories=CATEGORIES)


category_opt = cfg.SubCommandOpt('category',
                                 title='Command categories',
                                 help='Available categories',
                                 handler=add_command_parsers)


def main():
    """Parse options and call the appropriate class/method."""
    CONF.register_cli_opt(category_opt)
    config.parse_args(sys.argv)

    if CONF.category.name == "version":
        print(version.version_string_with_package())
        return 0

    if CONF.category.name == "bash-completion":
        cmd_common.print_bash_completion(CATEGORIES)
        return 0

    try:
        fn, fn_args, fn_kwargs = cmd_common.get_action_fn()
        ret = fn(*fn_args, **fn_kwargs)
        return ret
    except Exception:
        print(_('Error:\n%s') % traceback.format_exc())
        # This is 255 so it's not confused with the upgrade check exit codes.
        return 255



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from nova.policies import base


BASE_POLICY_NAME = 'os_compute_api:os-attach-interfaces'
POLICY_ROOT = 'os_compute_api:os-attach-interfaces:%s'

DEPRECATED_REASON = """
Nova API policies are introducing new default roles with scope_type
capabilities. Old policies are deprecated and silently going to be ignored
in nova 23.0.0 release.
"""

DEPRECATED_INTERFACES_POLICY = policy.DeprecatedRule(
    BASE_POLICY_NAME,
    base.RULE_ADMIN_OR_OWNER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='21.0.0',
)

attach_interfaces_policies = [
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'list',
        check_str=base.PROJECT_READER_OR_ADMIN,
        description="List port interfaces attached to a server",
        operations=[
            {
                'method': 'GET',
                'path': '/servers/{server_id}/os-interface'
            },
        ],
        scope_types=['project'],
        deprecated_rule=DEPRECATED_INTERFACES_POLICY),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'show',
        check_str=base.PROJECT_READER_OR_ADMIN,
        description="Show details of a port interface attached to a server",
        operations=[
            {
                'method': 'GET',
                'path': '/servers/{server_id}/os-interface/{port_id}'
            }
        ],
        scope_types=['project'],
        deprecated_rule=DEPRECATED_INTERFACES_POLICY),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'create',
        check_str=base.PROJECT_MEMBER_OR_ADMIN,
        description="Attach an interface to a server",
        operations=[
            {
                'method': 'POST',
                'path': '/servers/{server_id}/os-interface'
            }
        ],
        scope_types=['project'],
        deprecated_rule=DEPRECATED_INTERFACES_POLICY),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'delete',
        check_str=base.PROJECT_MEMBER_OR_ADMIN,
        description="Detach an interface from a server",
        operations=[
            {
                'method': 'DELETE',
                'path': '/servers/{server_id}/os-interface/{port_id}'
            }
        ],
        scope_types=['project'],
        deprecated_rule=DEPRECATED_INTERFACES_POLICY)
]


def list_rules():
    return attach_interfaces_policies



# Copyright 2012 Nebula, Inc.
# Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_utils.fixture import uuidsentinel as uuids

from nova.tests.functional.api import client as api_client
from nova.tests.functional import api_samples_test_base


class VirtualInterfacesJsonTest(api_samples_test_base.ApiSampleTestBase):
    api_major_version = 'v2'

    def test_vifs_list(self):
        uuid = uuids.instance_1

        ex = self.assertRaises(api_client.OpenStackApiException,
                               self.api.api_get,
                               '/servers/%s/os-virtual-interfaces' % uuid)
        self.assertEqual(410, ex.response.status_code)



# Copyright 2014 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova.api.validation import parameter_types

create = {
    'type': 'object',
    'properties': {
        'security_group': {
            'type': 'object',
            'properties': {
                'name': {
                    'type': 'string',
                    'minLength': 0,
                    'maxLength': 255,
                },
                'description': {
                    'type': 'string',
                    'minLength': 0,
                    'maxLength': 255,
                },
            },
            'required': ['name', 'description'],
            # NOTE(stephenfin): Per gmann's note below
            'additionalProperties': True,
        },
    },
    'required': ['security_group'],
    # NOTE(stephenfin): Per gmann's note below
    'additionalProperties': True,
}

update = create

create_rules = {
    'type': 'object',
    'properties': {
        'security_group_rule': {
            'type': 'object',
            'properties': {
                'group_id': {
                    'oneOf': [
                        {'type': 'null'},
                        {'type': 'string', 'format': 'uuid'},
                    ],
                },
                'parent_group_id': {
                    'type': 'string',
                    'format': 'uuid',
                },
                # NOTE(stephenfin): We never validated these and we're not
                # going to add that validation now.
                'to_port': {},
                'from_port': {},
                'ip_protocol': {},
                'cidr': {},
            },
            'required': ['parent_group_id'],
            # NOTE(stephenfin): Per gmann's note below
            'additionalProperties': True,
        },
    },
    'required': ['security_group_rule'],
    # NOTE(stephenfin): Per gmann's note below
    'additionalProperties': True,

}

# TODO(stephenfin): Remove additionalProperties in a future API version
show_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

index_query = {
    'type': 'object',
    'properties': {
        'limit': parameter_types.multi_params(
             parameter_types.non_negative_integer),
        'offset': parameter_types.multi_params(
             parameter_types.non_negative_integer),
        'all_tenants': parameter_types.multi_params({'type': 'string'})
    },
    # NOTE(gmann): This is kept True to keep backward compatibility.
    # As of now Schema validation stripped out the additional parameters and
    # does not raise 400. This API is deprecated in microversion 2.36 so we
    # do not to update the additionalProperties to False.
    'additionalProperties': True
}

# TODO(stephenfin): Remove additionalProperties in a future API version
server_sg_index_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

# TODO(stephenfin): Remove additionalProperties in a future API version
add_security_group = {
    'type': 'object',
    'properties': {
        'addSecurityGroup': {
            'type': 'object',
            'properties': {
                'name': {
                    'type': 'string',
                    'minLength': 1,
                },
            },
            'required': ['name'],
            'additionalProperties': False
        }
    },
    'required': ['addSecurityGroup'],
    'additionalProperties': True,
}

# TODO(stephenfin): Remove additionalProperties in a future API version
remove_security_group = {
    'type': 'object',
    'properties': {
        'removeSecurityGroup': {
            'type': 'object',
            'properties': {
                'name': {
                    'type': 'string',
                    'minLength': 1,
                },
            },
            'required': ['name'],
            'additionalProperties': False
        }
    },
    'required': ['removeSecurityGroup'],
    'additionalProperties': True,
}

add_security_group_response = {
    'type': 'null',
}

remove_security_group_response = {
    'type': 'null',
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from webob import exc

from nova.api.openstack.compute.schemas import networks_associate as schema
from nova.api.openstack import wsgi
from nova.api import validation


_removal_reason = """\
This action only works with *nova-network*, which was deprecated in the
14.0.0 (Newton) release.
It fails with HTTP 404 starting from microversion 2.36.
It was removed in the 21.0.0 (Ussuri) release.
"""


class NetworkAssociateActionController(wsgi.Controller):
    """Network Association API Controller."""

    @wsgi.action("disassociate_host")
    @wsgi.expected_errors(410)
    @wsgi.removed('18.0.0', _removal_reason)
    @validation.schema(schema.disassociate)
    def _disassociate_host_only(self, req, id, body):
        raise exc.HTTPGone()

    @wsgi.action("disassociate_project")
    @wsgi.expected_errors(410)
    @wsgi.removed('18.0.0', _removal_reason)
    @validation.schema(schema.disassociate_project)
    def _disassociate_project_only(self, req, id, body):
        raise exc.HTTPGone()

    @wsgi.action("associate_host")
    @wsgi.expected_errors(410)
    @wsgi.removed('18.0.0', _removal_reason)
    @validation.schema(schema.associate_host)
    def _associate_host(self, req, id, body):
        raise exc.HTTPGone()



# Copyright 2014 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from nova.api.validation import parameter_types

_availability_zone = {'oneOf': [parameter_types.az_name, {'type': 'null'}]}
_availability_zone_with_leading_trailing_spaces = {
    'oneOf': [parameter_types.az_name_with_leading_trailing_spaces,
              {'type': 'null'}]
}

create = {
    'type': 'object',
    'properties': {
        'aggregate': {
            'type': 'object',
            'properties': {
                'name': parameter_types.name,
                'availability_zone': _availability_zone,
            },
            'required': ['name'],
            'additionalProperties': False,
        },
    },
    'required': ['aggregate'],
    'additionalProperties': False,
}


create_v20 = copy.deepcopy(create)
create_v20['properties']['aggregate']['properties']['name'] = (
    parameter_types.name_with_leading_trailing_spaces)
create_v20['properties']['aggregate']['properties']['availability_zone'] = (
    _availability_zone_with_leading_trailing_spaces)


update = {
    'type': 'object',
    'properties': {
        'aggregate': {
            'type': 'object',
            'properties': {
                'name': parameter_types.name_with_leading_trailing_spaces,
                'availability_zone': _availability_zone
            },
            'additionalProperties': False,
            'anyOf': [
                {'required': ['name']},
                {'required': ['availability_zone']}
            ]
        },
    },
    'required': ['aggregate'],
    'additionalProperties': False,
}


update_v20 = copy.deepcopy(update)
update_v20['properties']['aggregate']['properties']['name'] = (
    parameter_types.name_with_leading_trailing_spaces)
update_v20['properties']['aggregate']['properties']['availability_zone'] = (
    _availability_zone_with_leading_trailing_spaces)


add_host = {
    'type': 'object',
    'properties': {
        'add_host': {
            'type': 'object',
            'properties': {
                'host': parameter_types.fqdn,
            },
            'required': ['host'],
            'additionalProperties': False,
        },
    },
    'required': ['add_host'],
    'additionalProperties': False,
}

remove_host = {
    'type': 'object',
    'properties': {
        'remove_host': {
            'type': 'object',
            'properties': {
                'host': parameter_types.fqdn,
            },
            'required': ['host'],
            'additionalProperties': False,
        },
    },
    'required': ['remove_host'],
    'additionalProperties': False,
}

set_metadata = {
    'type': 'object',
    'properties': {
        'set_metadata': {
            'type': 'object',
            'properties': {
                'metadata': parameter_types.metadata_with_null
            },
            'required': ['metadata'],
            'additionalProperties': False,
        },
    },
    'required': ['set_metadata'],
    'additionalProperties': False,
}

# TODO(stephenfin): Remove additionalProperties in a future API version
index_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

# TODO(stephenfin): Remove additionalProperties in a future API version
show_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

_aggregate_response = {
    'type': 'object',
    'properties': {
        'availability_zone': {'type': ['null', 'string']},
        'created_at': {'type': 'string', 'format': 'date-time'},
        'deleted': {'type': 'boolean'},
        'deleted_at': {
            'oneOf': [
                {'type': 'null'},
                {'type': 'string', 'format': 'date-time'},
            ],
        },
        'hosts': {
            'type': ['array', 'null'],
            'items': {
                'type': 'string',
            },
        },
        'id': {'type': 'integer'},
        # TODO(stephenfin): This could be stricter
        'metadata': {
            'type': ['null', 'object'],
            'properties': {},
            'additionalProperties': True,
        },
        'name': {'type': 'string'},
        'updated_at': {
            'oneOf': [
                {'type': 'null'},
                {'type': 'string', 'format': 'date-time'},
            ],
        },
    },
    'required': [
        'availability_zone',
        'created_at',
        'deleted',
        'deleted_at',
        'hosts',
        'id',
        'metadata',
        'name',
        'updated_at',
    ],
    'additionalProperties': False,
}

_aggregate_response_v241 = copy.deepcopy(_aggregate_response)
_aggregate_response_v241['properties'].update(
    {'uuid': {'type': 'string', 'format': 'uuid'}},
)
_aggregate_response_v241['required'].append('uuid')

index_response = {
    'type': 'object',
    'properties': {
        'aggregates': {
            'type': 'array',
            'items': copy.deepcopy(_aggregate_response),
        },
    },
    'required': ['aggregates'],
    'additionalProperties': False,
}
index_response_v241 = copy.deepcopy(index_response)
index_response_v241['properties']['aggregates']['items'] = copy.deepcopy(
    _aggregate_response_v241
)

create_response = {
    'type': 'object',
    'properties': {
        'aggregate': copy.deepcopy(_aggregate_response),
    },
    'required': ['aggregate'],
    'additionalProperties': False,
}
del create_response['properties']['aggregate']['properties']['hosts']
del create_response['properties']['aggregate']['properties']['metadata']
create_response['properties']['aggregate']['required'].remove('hosts')
create_response['properties']['aggregate']['required'].remove('metadata')
create_response_v241 = copy.deepcopy(create_response)
create_response_v241['properties']['aggregate'] = copy.deepcopy(
    _aggregate_response_v241
)
del create_response_v241['properties']['aggregate']['properties']['hosts']
del create_response_v241['properties']['aggregate']['properties']['metadata']
create_response_v241['properties']['aggregate']['required'].remove('hosts')
create_response_v241['properties']['aggregate']['required'].remove('metadata')

show_response = {
    'type': 'object',
    'properties': {
        'aggregate': copy.deepcopy(_aggregate_response),
    },
    'required': ['aggregate'],
    'additionalProperties': False,
}
show_response_v241 = copy.deepcopy(show_response)
show_response_v241['properties']['aggregate'] = copy.deepcopy(
    _aggregate_response_v241
)

update_response = copy.deepcopy(show_response)
update_response_v241 = copy.deepcopy(show_response_v241)

delete_response = {'type': 'null'}

add_host_response = copy.deepcopy(show_response)
add_host_response_v241 = copy.deepcopy(show_response_v241)

remove_host_response = copy.deepcopy(show_response)
remove_host_response_v241 = copy.deepcopy(show_response_v241)

set_metadata_response = copy.deepcopy(show_response)
set_metadata_response_v241 = copy.deepcopy(show_response_v241)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class VendorDataDriver(object):
    """The base VendorData Drivers should inherit from."""

    def __init__(self, *args, **kwargs):
        """Init method should do all expensive operations."""
        self._data = {}

    def get(self):
        """Return a dictionary of primitives to be rendered in metadata

        :return: A dictionary of primitives.
        """
        return self._data



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add_tls_port_to_console_auth_tokens

Revision ID: 2903cd72dc14
Revises: d60bddf7a903
Create Date: 2024-07-18 22:55:25.736157
"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '2903cd72dc14'
down_revision = 'd60bddf7a903'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('console_auth_tokens', schema=None) as batch_op:
        batch_op.add_column(sa.Column('tls_port', sa.Integer()))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_limit import fixture as limit_fixture
from oslo_serialization import base64
from oslo_utils.fixture import uuidsentinel as uuids

from nova import context as nova_context
from nova.limit import local as local_limit
from nova.objects import flavor as flavor_obj
from nova.objects import instance_group as group_obj
from nova.tests import fixtures as nova_fixtures
from nova.tests.functional.api import client
from nova.tests.functional import integrated_helpers


class UnifiedLimitsTest(integrated_helpers._IntegratedTestBase):

    def setUp(self):
        super(UnifiedLimitsTest, self).setUp()
        # Use different project_ids for non-admin and admin.
        self.api.project_id = 'fake'
        self.admin_api.project_id = 'admin'

        self.flags(driver="nova.quota.UnifiedLimitsDriver", group="quota")
        reglimits = {local_limit.SERVER_METADATA_ITEMS: 128,
                     local_limit.INJECTED_FILES: 5,
                     local_limit.INJECTED_FILES_CONTENT: 10 * 1024,
                     local_limit.INJECTED_FILES_PATH: 255,
                     local_limit.KEY_PAIRS: 100,
                     local_limit.SERVER_GROUPS: 10,
                     local_limit.SERVER_GROUP_MEMBERS: 1,
                     'servers': 4,
                     'class:VCPU': 8,
                     'class:MEMORY_MB': 32768,
                     'class:DISK_GB': 250}
        projlimits = {self.api.project_id: {'servers': 2,
                                            'class:VCPU': 4,
                                            'class:MEMORY_MB': 16384,
                                            'class:DISK_GB': 100}}
        self.useFixture(limit_fixture.LimitFixture(reglimits, projlimits))
        self.ctx = nova_context.get_admin_context()

    def _setup_services(self):
        # Use driver with lots of resources so we don't get NoValidHost while
        # testing quotas. Need to do this before services are started.
        self.flags(compute_driver='fake.FakeDriver')
        super(UnifiedLimitsTest, self)._setup_services()

    def test_servers(self):
        # First test the project limit using the non-admin project.
        for i in range(2):
            self._create_server(api=self.api)

        # Attempt to create a third server should fail.
        e = self.assertRaises(
            client.OpenStackApiException, self._create_server, api=self.api)
        self.assertEqual(403, e.response.status_code)
        self.assertIn('servers', e.response.text)

        # Then test the default limit using the admin project.
        for i in range(4):
            self._create_server(api=self.admin_api)

        # Attempt to create a fifth server should fail.
        e = self.assertRaises(
            client.OpenStackApiException, self._create_server,
            api=self.admin_api)
        self.assertEqual(403, e.response.status_code)
        self.assertIn('servers', e.response.text)

    def test_vcpu(self):
        # First test the project limit using the non-admin project.
        # m1.large has vcpus=4 and our project limit is 4, should succeed.
        flavor = flavor_obj.Flavor.get_by_name(self.ctx, 'm1.large')
        self._create_server(api=self.api, flavor_id=flavor.flavorid)

        # m1.small has vcpus=1, should fail because we are at quota.
        flavor = flavor_obj.Flavor.get_by_name(self.ctx, 'm1.small')
        e = self.assertRaises(
            client.OpenStackApiException, self._create_server, api=self.api,
            flavor_id=flavor.flavorid)
        self.assertEqual(403, e.response.status_code)
        self.assertIn('class:VCPU', e.response.text)

        # Then test the default limit of 8 using the admin project.
        flavor = flavor_obj.Flavor.get_by_name(self.ctx, 'm1.large')
        for i in range(2):
            self._create_server(api=self.admin_api, flavor_id=flavor.flavorid)

        # Attempt to create another server with vcpus=1 should fail because we
        # are at quota.
        flavor = flavor_obj.Flavor.get_by_name(self.ctx, 'm1.small')
        e = self.assertRaises(
            client.OpenStackApiException, self._create_server,
            api=self.admin_api, flavor_id=flavor.flavorid)
        self.assertEqual(403, e.response.status_code)
        self.assertIn('class:VCPU', e.response.text)

    def test_memory_mb(self):
        # First test the project limit using the non-admin project.
        flavor = flavor_obj.Flavor(
            context=self.ctx, memory_mb=16384, vcpus=1, root_gb=1,
            flavorid='9', name='m1.custom')
        flavor.create()
        self._create_server(api=self.api, flavor_id=flavor.flavorid)

        # Attempt to create another should fail as we are at quota.
        e = self.assertRaises(
            client.OpenStackApiException, self._create_server, api=self.api,
            flavor_id=flavor.flavorid)
        self.assertEqual(403, e.response.status_code)
        self.assertIn('class:MEMORY_MB', e.response.text)

        # Then test the default limit of 32768 using the admin project.
        for i in range(2):
            self._create_server(api=self.admin_api, flavor_id=flavor.flavorid)

        # Attempt to create another server should fail because we are at quota.
        e = self.assertRaises(
            client.OpenStackApiException, self._create_server,
            api=self.admin_api, flavor_id=flavor.flavorid)
        self.assertEqual(403, e.response.status_code)
        self.assertIn('class:MEMORY_MB', e.response.text)

    def test_disk_gb(self):
        # First test the project limit using the non-admin project.
        flavor = flavor_obj.Flavor(
            context=self.ctx, memory_mb=1, vcpus=1, root_gb=100,
            flavorid='9', name='m1.custom')
        flavor.create()
        self._create_server(api=self.api, flavor_id=flavor.flavorid)

        # Attempt to create another should fail as we are at quota.
        e = self.assertRaises(
            client.OpenStackApiException, self._create_server, api=self.api,
            flavor_id=flavor.flavorid)
        self.assertEqual(403, e.response.status_code)
        self.assertIn('class:DISK_GB', e.response.text)

        # Then test the default limit of 250 using the admin project.
        for i in range(2):
            self._create_server(api=self.admin_api, flavor_id=flavor.flavorid)

        # Attempt to create another server should fail because we are at quota.
        e = self.assertRaises(
            client.OpenStackApiException, self._create_server,
            api=self.admin_api, flavor_id=flavor.flavorid)
        self.assertEqual(403, e.response.status_code)
        self.assertIn('class:DISK_GB', e.response.text)

    def test_no_injected_files(self):
        self._create_server()

    def test_max_injected_files(self):
        # Quota is 5.
        files = []
        contents = base64.encode_as_text('some content')
        for i in range(5):
            files.append(('/my/path%d' % i, contents))
        server = self._build_server()
        personality = [
            {'path': item[0], 'contents': item[1]} for item in files]
        server['personality'] = personality
        server = self.api.post_server({'server': server})
        self.api.delete_server(server['id'])

    def test_max_injected_file_content_bytes(self):
        # Quota is 10 * 1024
        # Hm, apparently quota is checked against the base64 encoded string
        # even though the api-ref claims the limit is for the decoded data.
        # Subtract 3072 characters to account for that.
        content = base64.encode_as_bytes(
            ''.join(['a' for i in range(10 * 1024 - 3072)]))
        server = self._build_server()
        personality = [{'path': '/test/path', 'contents': content}]
        server['personality'] = personality
        server = self.api.post_server({'server': server})
        self.api.delete_server(server['id'])

    def test_max_injected_file_path_bytes(self):
        # Quota is 255.
        path = ''.join(['a' for i in range(255)])
        contents = base64.encode_as_text('some content')
        server = self._build_server()
        personality = [{'path': path, 'contents': contents}]
        server['personality'] = personality
        server = self.api.post_server({'server': server})
        self.api.delete_server(server['id'])

    def test_server_group_members(self):
        # Create a server group.
        instance_group = group_obj.InstanceGroup(
            self.ctx, policy="anti-affinity")
        instance_group.name = "foo"
        instance_group.project_id = self.ctx.project_id
        instance_group.user_id = self.ctx.user_id
        instance_group.uuid = uuids.instance_group
        instance_group.create()

        # Quota for server group members is 1.
        server = self._build_server()
        hints = {'group': uuids.instance_group}
        req = {'server': server, 'os:scheduler_hints': hints}
        server = self.admin_api.post_server(req)

        # Attempt to create another server in the group should fail because we
        # are at quota.
        e = self.assertRaises(
            client.OpenStackApiException, self.admin_api.post_server, req)
        self.assertEqual(403, e.response.status_code)
        self.assertIn('server_group_members', e.response.text)

        self.admin_api.delete_server(server['id'])


class ResourceStrategyTest(integrated_helpers._IntegratedTestBase):

    def setUp(self):
        super().setUp()
        # Use different project_ids for non-admin and admin.
        self.api.project_id = 'fake'
        self.admin_api.project_id = 'admin'

        self.flags(driver="nova.quota.UnifiedLimitsDriver", group='quota')
        self.ctx = nova_context.get_admin_context()
        self.ul_api = self.useFixture(nova_fixtures.UnifiedLimitsFixture())

    def test_invalid_value_in_resource_list(self):
        # First two have casing issues, next doesn't have the "class:" prefix,
        # last is a typo.
        invalid_names = (
            'class:vcpu', 'class:CUSTOM_thing', 'VGPU', 'class:MEMRY_MB')
        for name in invalid_names:
            e = self.assertRaises(
                ValueError, self.flags, unified_limits_resource_list=[name],
                group='quota')
            self.assertIsInstance(e, ValueError)
            self.assertIn('not a valid resource class name', str(e))

    def test_valid_custom_resource_classes(self):
        valid_names = ('class:CUSTOM_GOLD', 'class:CUSTOM_A5_1')
        for name in valid_names:
            self.flags(unified_limits_resource_list=[name], group='quota')

    @mock.patch('nova.limit.utils.LOG.error')
    def test_invalid_strategy_configuration(self, mock_log_error):
        # Quota should be enforced and fail the check if there is somehow an
        # invalid strategy value.
        self.stub_out(
            'nova.limit.utils.CONF.quota.unified_limits_resource_strategy',
            'bogus')
        e = self.assertRaises(
            client.OpenStackApiException, self._create_server)
        self.assertEqual(403, e.response.status_code)
        expected = (
            'Invalid strategy value: bogus is specified in the '
            '[quota]unified_limits_resource_strategy config option, so '
            'enforcing for resources')
        mock_log_error.assert_called()
        self.assertIn(expected, mock_log_error.call_args.args[0])

    @mock.patch('nova.limit.utils.LOG.debug')
    def test_required_limits_set(self, mock_log_debug):
        # Required resources configured are: 'servers', MEMORY_MB, and VCPU
        self.flags(unified_limits_resource_strategy='require', group='quota')
        require = ['servers', 'class:MEMORY_MB', 'class:VCPU']
        self.flags(unified_limits_resource_list=require, group='quota')
        # Purposely not setting any quota for DISK_GB.
        self.ul_api.create_registered_limit(
            resource_name='servers', default_limit=4)
        self.ul_api.create_registered_limit(
            resource_name='class:VCPU', default_limit=8)
        self.ul_api.create_registered_limit(
            resource_name='class:MEMORY_MB', default_limit=32768)
        # Server create should succeed because required resources VCPU,
        # MEMORY_MB, and 'servers' have registered limits.
        self._create_server()
        unset_limits = set(['class:DISK_GB'])
        call = mock.call(
            f'Resources {unset_limits} have no registered limits set in '
            f'Keystone. [quota]unified_limits_resource_strategy is require '
            f'and [quota]unified_limits_resource_list is {require}, so not '
            'enforcing')
        # The message will be logged twice -- once in nova-api and once in
        # nova-conductor because of the quota recheck after resource creation.
        self.assertEqual([call, call], mock_log_debug.mock_calls)

    def test_some_required_limits_not_set(self):
        # Now add DISK_GB as a required resource.
        self.flags(unified_limits_resource_strategy='require', group='quota')
        self.flags(unified_limits_resource_list=[
            'servers', 'class:MEMORY_MB', 'class:VCPU', 'class:DISK_GB'],
            group='quota')
        # Purposely not setting any quota for DISK_GB.
        self.ul_api.create_registered_limit(
            resource_name='servers', default_limit=-1)
        self.ul_api.create_registered_limit(
            resource_name='class:VCPU', default_limit=8)
        self.ul_api.create_registered_limit(
            resource_name='class:MEMORY_MB', default_limit=32768)
        # Server create should fail because required resource DISK_GB does not
        # have a registered limit set.
        e = self.assertRaises(
            client.OpenStackApiException, self._create_server, api=self.api)
        self.assertEqual(403, e.response.status_code)

    @mock.patch('nova.limit.utils.LOG.debug')
    def test_no_required_limits(self, mock_log_debug):
        # Configured to not require any resource limits.
        self.flags(unified_limits_resource_strategy='require', group='quota')
        self.flags(unified_limits_resource_list=[], group='quota')
        # Server create should succeed because no resource registered limits
        # are required to be set.
        self._create_server()
        # The message will be logged twice -- once in nova-api and once in
        # nova-conductor because of the quota recheck after resource creation.
        self.assertEqual(2, mock_log_debug.call_count)

    @mock.patch('nova.limit.utils.LOG.debug')
    def test_ignored_limits_set(self, mock_log_debug):
        # Ignored unset limit resources configured is DISK_GB.
        self.flags(unified_limits_resource_strategy='ignore', group='quota')
        ignore = ['class:DISK_GB']
        self.flags(unified_limits_resource_list=ignore, group='quota')
        # Purposely not setting any quota for DISK_GB.
        self.ul_api.create_registered_limit(
            resource_name='servers', default_limit=4)
        self.ul_api.create_registered_limit(
            resource_name='class:VCPU', default_limit=8)
        self.ul_api.create_registered_limit(
            resource_name='class:MEMORY_MB', default_limit=32768)
        # Server create should succeed because class:DISK_GB is specified in
        # the ignore unset limit list.
        self._create_server()
        unset_limits = set(['class:DISK_GB'])
        call = mock.call(
            f'Resources {unset_limits} have no registered limits set in '
            f'Keystone. [quota]unified_limits_resource_strategy is ignore and '
            f'[quota]unified_limits_resource_list is {ignore}, so not '
            'enforcing')
        # The message will be logged twice -- once in nova-api and once in
        # nova-conductor because of the quota recheck after resource creation.
        self.assertEqual([call, call], mock_log_debug.mock_calls)

    def test_some_ignored_limits_not_set(self):
        # Configured to ignore only one unset resource limit.
        self.flags(unified_limits_resource_strategy='ignore', group='quota')
        self.flags(unified_limits_resource_list=[
            'class:DISK_GB'], group='quota')
        # Purposely not setting any quota for servers.
        self.ul_api.create_registered_limit(
            resource_name='class:VCPU', default_limit=8)
        self.ul_api.create_registered_limit(
            resource_name='class:MEMORY_MB', default_limit=32768)
        # Server create should fail because although resource DISK_GB does not
        # have a registered limit set and it is in the ignore list, resource
        # 'servers' does not have a limit set and it is not in the ignore list.
        e = self.assertRaises(
            client.OpenStackApiException, self._create_server, api=self.api)
        self.assertEqual(403, e.response.status_code)

    def test_no_ignored_limits(self):
        # Configured to not ignore any unset resource limits.
        self.flags(unified_limits_resource_strategy='ignore', group='quota')
        self.flags(unified_limits_resource_list=[], group='quota')
        # Server create should fail because resource DISK_GB does not have a
        # registered limit set and it is not in the ignore list.
        e = self.assertRaises(
            client.OpenStackApiException, self._create_server, api=self.api)
        self.assertEqual(403, e.response.status_code)

    def test_all_unlimited(self):
        # -1 is documented in Keystone as meaning unlimited:
        #
        # https://docs.openstack.org/keystone/latest/admin/unified-limits.html#what-is-a-limit
        #
        # but oslo.limit enforce does not treat -1 as unlimited at this time
        # and instead uses its literal integer value.
        #
        # Test that we consider -1 to be unlimited in Nova and the server
        # create should succeed.
        for resource in (
                'servers', 'class:VCPU', 'class:MEMORY_MB', 'class:DISK_GB'):
            self.ul_api.create_registered_limit(
                resource_name=resource, default_limit=-1)
        self._create_server()

    def test_default_unlimited_but_project_limited(self):
        # If the default limit is set to -1 unlimited but the project has a
        # limit, quota should be enforced at the project level.
        # Note that it is not valid to set a project limit without first
        # setting a registered limit -- Keystone will not allow it.
        self.ul_api.create_registered_limit(
            resource_name='servers', default_limit=-1)
        self.ul_api.create_limit(
            project_id='fake', resource_name='servers', resource_limit=1)
        # First server should succeed because we have a project limit of 1.
        self._create_server()
        # Second server should fail because it would exceed the project limit
        # of 1.
        e = self.assertRaises(
            client.OpenStackApiException, self._create_server, api=self.api)
        self.assertEqual(403, e.response.status_code)

    def test_default_limited_but_project_unlimited(self):
        # If the default limit is set to a value but the project has a limit
        # set to -1 unlimited, quota should be enforced at the project level.
        # Note that it is not valid to set a project limit without first
        # setting a registered limit -- Keystone will not allow it.
        self.ul_api.create_registered_limit(
            resource_name='servers', default_limit=0)
        self.ul_api.create_limit(
            project_id='fake', resource_name='servers', resource_limit=-1)
        # First server should succeed because we have a default limit of 0 and
        # a project limit of -1 unlimited.
        self._create_server()
        # Try to create a server in a different project (admin project) -- this
        # should fail because the default limit has been explicitly set to 0.
        e = self.assertRaises(
            client.OpenStackApiException, self._create_server,
            api=self.admin_api)
        self.assertEqual(403, e.response.status_code)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# TODO(stephenfin): Remove additionalProperties in a future API version
index_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

# TODO(stephenfin): Remove additionalProperties in a future API version
show_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

_extension_obj = {
    'type': 'object',
    'properties': {
        'alias': {'type': 'string'},
        'description': {'type': 'string'},
        'links': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'href': {
                        'type': 'string',
                        'format': 'url'
                    },
                    'rel': {
                        'type': 'string'
                    },
                }
            },
        },
        'name': {'type': 'string'},
        'namespace': {'type': 'string'},
        'updated': {'type': 'string', 'format': 'date-time'},
    },
    'required': [
        'alias', 'description', 'links', 'name', 'namespace', 'updated'
    ],
    'additionalProperties': False,
}


index_response = {
    'type': 'object',
    'properties': {
        'extensions': {
            'type': 'array',
            'items': _extension_obj,
        }
    },
    'required': ['extensions'],
    'additionalProperties': False,
}

show_response = {
    'type': 'object',
    'properties': {
        'extension': _extension_obj,
    },
    'required': ['extension'],
    'additionalProperties': False,
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova.api.openstack import common
from nova.api.openstack.compute.views import servers


class ViewBuilder(common.ViewBuilder):
    _collection_name = 'shares'

    def __init__(self):
        super(ViewBuilder, self).__init__()
        self._server_builder = servers.ViewBuilder()

    def _list_view(self, db_shares):
        shares = {'shares': []}
        for db_share in db_shares:
            share = {
                'share_id': db_share.share_id,
                'status': db_share.status,
                'tag': db_share.tag,
            }
            shares['shares'].append(share)
        return shares

    def _show_view(self, context, db_share):
        share = {'share': {
            'share_id': db_share.share_id,
            'status': db_share.status,
            'tag': db_share.tag,
        }}

        if context.is_admin:
            share['share']['export_location'] = db_share.export_location
            share['share']['uuid'] = db_share.uuid

        return share



# Copyright (c) 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import webob.exc

from nova.api.openstack.compute.schemas import certificates as schema
from nova.api.openstack import wsgi
from nova.api import validation

_removal_reason = """\
This API was only used to build euca bundles when Nova had an in-tree EC2 API.
It no longer interacts with any parts of the system besides its own certificate
daemon.
It was removed in the 16.0.0 (Pike) release.
"""


class CertificatesController(wsgi.Controller):
    """The x509 Certificates API controller for the OpenStack API."""

    @wsgi.expected_errors(410)
    @wsgi.removed('16.0.0', _removal_reason)
    @validation.query_schema(schema.show_query)
    def show(self, req, id):
        """Return certificate information."""
        raise webob.exc.HTTPGone()

    @wsgi.expected_errors((410))
    @wsgi.removed('16.0.0', _removal_reason)
    @validation.schema(schema.create)
    def create(self, req, body=None):
        """Create a certificate."""
        raise webob.exc.HTTPGone()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import fixtures
from oslo_utils.fixture import uuidsentinel as uuids
from oslo_utils import timeutils

from nova.api.openstack.compute import admin_password
from nova.compute import vm_states
from nova import exception
from nova.policies import admin_password as ap_policies
from nova.tests.unit.api.openstack import fakes
from nova.tests.unit import fake_instance
from nova.tests.unit.policies import base


class AdminPasswordPolicyTest(base.BasePolicyTest):
    """Test Admin Password APIs policies with all possible context.
    This class defines the set of context with different roles
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will call the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(AdminPasswordPolicyTest, self).setUp()
        self.controller = admin_password.AdminPasswordController()
        self.req = fakes.HTTPRequest.blank('')
        user_id = self.req.environ['nova.context'].user_id
        self.rule_name = ap_policies.BASE_POLICY_NAME
        self.mock_get = self.useFixture(
            fixtures.MockPatch('nova.api.openstack.common.get_instance')).mock
        uuid = uuids.fake_id
        self.instance = fake_instance.fake_instance_obj(
                self.project_member_context,
                id=1, uuid=uuid, project_id=self.project_id,
                user_id=user_id, vm_state=vm_states.ACTIVE,
                task_state=None, launched_at=timeutils.utcnow())
        self.mock_get.return_value = self.instance
        # With legacy rule and no scope checks, all admin, project members
        # project reader or other project role(because legacy rule allow server
        # owner- having same project id and no role check) is able to change
        # the password for their server.
        self.project_action_authorized_contexts = [
            self.legacy_admin_context, self.system_admin_context,
            self.project_admin_context, self.project_member_context,
            self.project_reader_context, self.project_foo_context]

    @mock.patch('nova.compute.api.API.set_admin_password')
    def test_change_paassword_policy(self, mock_password):
        self.common_policy_auth(self.project_action_authorized_contexts,
                                self.rule_name,
                                self.controller.change_password,
                                self.req, self.instance.uuid,
                                body={'changePassword': {
                                      'adminPass': '1234pass'}})

    def test_change_password_overridden_policy_failed_with_other_user(self):
        # Change the user_id in request context.
        req = fakes.HTTPRequest.blank('')
        req.environ['nova.context'].user_id = 'other-user'
        body = {'changePassword': {'adminPass': '1234pass'}}
        self.policy.set_rules({self.rule_name: "user_id:%(user_id)s"})
        exc = self.assertRaises(
            exception.PolicyNotAuthorized, self.controller.change_password,
            req, fakes.FAKE_UUID, body=body)
        self.assertEqual(
            "Policy doesn't allow %s to be performed." % self.rule_name,
            exc.format_message())

    @mock.patch('nova.compute.api.API.set_admin_password')
    def test_change_password_overridden_policy_pass_with_same_user(
        self, password_mock):
        self.policy.set_rules({self.rule_name: "user_id:%(user_id)s"})
        body = {'changePassword': {'adminPass': '1234pass'}}
        self.controller.change_password(self.req, fakes.FAKE_UUID, body=body)
        password_mock.assert_called_once_with(self.req.environ['nova.context'],
                                              mock.ANY, '1234pass')


class AdminPasswordNoLegacyNoScopePolicyTest(AdminPasswordPolicyTest):
    """Test Admin Password APIs policies with no legacy deprecated rules
    and no scope checks which means new defaults only.

    """

    without_deprecated_rules = True

    def setUp(self):
        super(AdminPasswordNoLegacyNoScopePolicyTest, self).setUp()
        # With no legacy rule, only project admin or member will be
        # able to change the server password.
        self.project_action_authorized_contexts = (
            self.project_member_or_admin_with_no_scope_no_legacy)


class AdminPasswordScopeTypePolicyTest(AdminPasswordPolicyTest):
    """Test Admin Password APIs policies with system scope enabled.
    This class set the nova.conf [oslo_policy] enforce_scope to True
    so that we can switch on the scope checking on oslo policy side.
    It defines the set of context with scoped token
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will run the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(AdminPasswordScopeTypePolicyTest, self).setUp()
        self.flags(enforce_scope=True, group="oslo_policy")
        # Scope enable will not allow system admin to change password.
        self.project_action_authorized_contexts = (
            self.project_m_r_or_admin_with_scope_and_legacy)


class AdminPasswordScopeTypeNoLegacyTest(AdminPasswordScopeTypePolicyTest):
    """Test Admin Password APIs policies with system scope enabled,
    and no more deprecated rules which means scope + new defaults so
    only project admin and member is able to change their server password.
    """

    without_deprecated_rules = True

    def setUp(self):
        super(AdminPasswordScopeTypeNoLegacyTest, self).setUp()
        self.flags(enforce_scope=True, group="oslo_policy")

        # With scope enable and no legacy rule only project admin/member
        # will be able to change password for the server.
        self.project_action_authorized_contexts = (
            self.project_member_or_admin_with_scope_no_legacy)



# SPDX-License-Identifier: Apache-2.0
import functools
import os
import shutil

from unittest import mock

import fixtures

from oslo_utils.fixture import uuidsentinel as uuids
from oslo_utils.imageutils import format_inspector
from oslo_utils import units

import nova.conf

from nova import objects
from nova import test
from nova.virt.libvirt import driver
from nova.virt.libvirt import imagebackend
from nova.virt.libvirt import utils as libvirt_utils


CONF = nova.conf.CONF


class TestBugBackingFilePartitionTables(test.NoDBTestCase):
    """Regression test for nova created backing files

    This test case is for a bug that was discovered where nova was creating
    backing files for swap and ephemeral disks. With the move to using the
    image format inspector form oslo.utils we gained the ability to inspect
    files for mbr and gpt partition tables.

    This is now a requirement for all backing files created by nova.
    """

    def setUp(self):
        super(TestBugBackingFilePartitionTables, self).setUp()
        self.base_dir = self.useFixture(
            fixtures.TempDir())
        if shutil.which("qemu-img") is None:
            self.skipTest("qemu-img not installed")
        if shutil.which("mkfs.vfat") is None:
            self.skipTest("mkfs.vfat not installed")

    def test_create_file(self):
        """Test that files created files have partition tables

        This test will create a file and then inspect it to ensure
        that it has a partition table so that it can be used as a
        backing file.
        """

        file_path = os.path.join(self.base_dir.path, 'test_file')
        libvirt_utils.create_image(file_path, 'raw', '64M')
        self.assertTrue(os.path.exists(file_path))
        # FIXME(sean-k-mooney): oslo currently detect vfat as an mbr partition
        self.assertRaises(
            format_inspector.ImageFormatError,
            format_inspector.GPTInspector.from_file, file_path)

        # nova files should pass the RawFileInspector safety check
        inspector = format_inspector.RawFileInspector.from_file(file_path)
        self.assertIsNotNone(inspector)
        inspector.safety_check()

    def test_cache_file(self):
        """Test the qcow2 cache interaction for ephemeral disks

        This test will create a file via the image backend cache function
        and ensure that the backing file has a partition table
        """
        _create_ephemeral = driver.LibvirtDriver._create_ephemeral

        self.flags(use_cow_images=True)
        self.flags(instances_path=self.base_dir.path)
        self.flags(group='libvirt', images_type='qcow2')

        backend_image = imagebackend.Backend(CONF.use_cow_images).backend()
        instance = objects.Instance(uuid=uuids.instance)
        image = backend_image(instance, 'test_image')

        fn = functools.partial(
            _create_ephemeral, fs_label='ephemeral0',
            os_type=None, is_block_dev=False)
        # this need to be multiples of 1G
        size = 1 * units.Gi
        fname = "ephemeral_%s_%s" % (size, ".img")
        with mock.patch.object(
            imagebackend, '_update_utime_ignore_eacces') as m:

            image.cache(
                fetch_func=fn, context=None, filename=fname,
                size=size, ephemeral_size=1, safe=True)
            m.assert_called_once()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova.api.openstack.compute import extension_info
from nova.policies import extensions as policies
from nova.tests.unit.api.openstack import fakes
from nova.tests.unit.policies import base


class ExtensionsPolicyTest(base.BasePolicyTest):
    """Test Extensions APIs policies with all possible context.
    This class defines the set of context with different roles
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will call the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(ExtensionsPolicyTest, self).setUp()
        self.controller = extension_info.ExtensionInfoController()
        self.req = fakes.HTTPRequest.blank('')

        # Check that everyone is able to get extension info.
        self.everyone_authorized_contexts = [
            self.legacy_admin_context, self.system_admin_context,
            self.project_admin_context, self.project_member_context,
            self.project_reader_context, self.project_foo_context,
            self.other_project_reader_context,
            self.system_member_context, self.system_reader_context,
            self.system_foo_context,
            self.other_project_member_context
        ]
        self.everyone_unauthorized_contexts = []

    def test_list_extensions_policy(self):
        rule_name = policies.BASE_POLICY_NAME
        self.common_policy_check(self.everyone_authorized_contexts,
                                 self.everyone_unauthorized_contexts,
                                 rule_name,
                                 self.controller.index,
                                 self.req)

    def test_show_extensions_policy(self):
        rule_name = policies.BASE_POLICY_NAME
        self.common_policy_check(self.everyone_authorized_contexts,
                                 self.everyone_unauthorized_contexts,
                                 rule_name,
                                 self.controller.show,
                                 self.req, 'os-volumes')


class ExtensionsScopeTypePolicyTest(ExtensionsPolicyTest):
    """Test Extensions APIs policies with system scope enabled.
    This class set the nova.conf [oslo_policy] enforce_scope to True
    so that we can switch on the scope checking on oslo policy side.
    It defines the set of context with scoped token
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will run the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(ExtensionsScopeTypePolicyTest, self).setUp()
        self.flags(enforce_scope=True, group="oslo_policy")
        self.everyone_authorized_contexts = [
            self.legacy_admin_context,
            self.project_admin_context, self.project_member_context,
            self.project_reader_context, self.project_foo_context,
            self.other_project_reader_context,
            self.other_project_member_context
        ]
        self.everyone_unauthorized_contexts = [
            self.system_admin_context, self.system_member_context,
            self.system_reader_context, self.system_foo_context]


class ExtensionsNoLegacyPolicyTest(ExtensionsScopeTypePolicyTest):
    """Test Extensions APIs policies with system scope enabled,
    and no more deprecated rules.
    """
    without_deprecated_rules = True



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# TODO(stephenfin): Remove additionalProperties in a future API version
show_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

# TODO(stephenfin): Remove additionalProperties in a future API version
multi_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

_version_obj = {
    'type': 'object',
    'properties': {
        'id': {'type': 'string'},
        'status': {
            'type': 'string',
            'enum': ['CURRENT', 'SUPPORTED', 'DEPRECATED'],
        },
        'links': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'href': {'type': 'string'},
                    'rel': {'type': 'string'},
                    'type': {'type': 'string'},
                },
                'required': ['rel', 'href'],
                'additionalProperties': False,
            },
        },
        'min_version': {'type': 'string'},
        'updated': {'type': 'string', 'format': 'date-time'},
        'version': {'type': 'string'},
    },
    'required': ['id', 'status', 'links', 'min_version', 'updated'],
    'additionalProperties': False,
}

index_response = {
    'type': 'object',
    'properties': {
        'versions': {'type': 'array', 'items': _version_obj}
    },
    'required': ['versions'],
    'additionalProperties': False,
}

_version_obj_with_media_types = _version_obj
_version_obj_with_media_types['properties'].update({
    'media-types': {
        'type': 'array',
        'items': {
            'type': 'object',
            'properties': {
                'base': {'type': 'string'},
                'type': {'type': 'string'},
            },
            'required': ['base', 'type'],
            'additionalProperties': False,
        },
    }
})

show_response = {
    'type': 'object',
    'properties': {
        'version': _version_obj_with_media_types
    },
    'required': ['version'],
    'additionalProperties': False,
}

_legacy_version_obj = {
    'type': 'object',
    'properties': {
        'id': {'type': 'string'},
        'status': {'type': 'string'},
        'links': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'href': {'type': 'string'},
                    'rel': {'type': 'string'},
                    'type': {'type': 'string'},
                },
                'required': ['rel', 'href'],
                'additionalProperties': False,
            },
        },
        'media-types': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'base': {'type': 'string'},
                    'type': {'type': 'string'},
                },
                'required': ['base', 'type'],
                'additionalProperties': False,
            },
        },
    },
    'required': ['id', 'status', 'links', 'media-types'],
    'additionalProperties': False,
}

multi_response = {
    'type': 'object',
    'properties': {
        'choices': {'type': 'array', 'items': _legacy_version_obj}
    },
    'required': ['choices'],
    'additionalProperties': False,
}



# Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
from oslo_utils.fixture import uuidsentinel as uuids
import webob

from nova.api.openstack.compute import suspend_server
from nova import exception
from nova.tests.unit.api.openstack.compute import admin_only_action_common


@ddt.ddt
class SuspendServerTests(admin_only_action_common.CommonTests):

    def setUp(self):
        super().setUp()
        self.controller = suspend_server.SuspendServerController()
        self.compute_api = self.controller.compute_api

    def test_suspend_resume(self):
        self._test_actions(
            ['_suspend', '_resume'],
            body_map={
                '_suspend': {'suspend': None}, '_resume': {'resume': None}
            }
        )

    @ddt.data(
        exception.OperationNotSupportedForVDPAInterface(
            instance_uuid=uuids.instance, operation='foo'),
        exception.OperationNotSupportedForSEV(
            instance_uuid=uuids.instance, operation='foo'),
        exception.ForbiddenWithAccelerators(),
    )
    @mock.patch('nova.compute.api.API.suspend')
    def test_suspend_raise_badrequest_for_not_supported_features(
        self, exc, mock_suspend):
        mock_suspend.side_effect = exc
        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller._suspend,
            self.req, uuids.instance, body={'suspend': None})
        self.assertTrue(mock_suspend.called)

    def test_suspend_resume_with_non_existed_instance(self):
        self._test_actions_with_non_existed_instance(
            ['_suspend', '_resume'],
            body_map={
                '_suspend': {'suspend': None}, '_resume': {'resume': None}
            }
        )

    def test_suspend_resume_raise_conflict_on_invalid_state(self):
        self._test_actions_raise_conflict_on_invalid_state(
            ['_suspend', '_resume'],
            body_map={
                '_suspend': {'suspend': None}, '_resume': {'resume': None}
            }
        )

    def test_actions_with_locked_instance(self):
        self._test_actions_with_locked_instance(
            ['_suspend', '_resume'],
            body_map={
                '_suspend': {'suspend': None}, '_resume': {'resume': None}
            }
        )



# Copyright 2017 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from nova.api.validation import parameter_types

# NOTE(takashin): The following sort keys are defined for backward
# compatibility. If they are changed, the API microversion should be bumped.
VALID_SORT_KEYS = [
    'created_at', 'description', 'disabled', 'ephemeral_gb', 'flavorid', 'id',
    'is_public', 'memory_mb', 'name', 'root_gb', 'rxtx_factor', 'swap',
    'updated_at', 'vcpu_weight', 'vcpus'
]

VALID_SORT_DIR = ['asc', 'desc']

create = {
    'type': 'object',
    'properties': {
        'flavor': {
            'type': 'object',
            'properties': {
                # in nova/flavors.py, name with all white spaces is forbidden.
                'name': parameter_types.name,
                # forbid leading/trailing whitespaces
                'id': {
                    'type': ['string', 'number', 'null'],
                    'minLength': 1, 'maxLength': 255,
                    'pattern': '^(?! )[a-zA-Z0-9. _-]+(?<! )$'
                },
                'ram': parameter_types.flavor_param_positive,
                'vcpus': parameter_types.flavor_param_positive,
                'disk': parameter_types.flavor_param_non_negative,
                'OS-FLV-EXT-DATA:ephemeral':
                    parameter_types.flavor_param_non_negative,
                'swap': parameter_types.flavor_param_non_negative,
                # positive ( > 0) float
                'rxtx_factor': {
                    'type': ['number', 'string'],
                    'pattern': r'^[0-9]+(\.[0-9]+)?$',
                    # this is a float, so we want to allow everything close to
                    # 0 (e.g. 0.1) but not 0 itself, hence exclusiveMinimum
                    # rather than the usual minimum
                    'exclusiveMinimum': 0,
                    # maximum's value is limited to db constant's
                    # SQL_SP_FLOAT_MAX (in nova/db/constants.py)
                    'maximum': 3.40282e+38
                },
                'os-flavor-access:is_public': parameter_types.boolean,
            },
            # TODO(oomichi): 'id' should be required with v2.1+microversions.
            # On v2.0 API, nova-api generates a flavor-id automatically if
            # specifying null as 'id' or not specifying 'id'. Ideally a client
            # should specify null as 'id' for requesting auto-generated id
            # exactly. However, this strict limitation causes a backwards
            # incompatible issue on v2.1. So now here relaxes the requirement
            # of 'id'.
            'required': ['name', 'ram', 'vcpus', 'disk'],
            'additionalProperties': False,
        },
    },
    'required': ['flavor'],
    'additionalProperties': False,
}


create_v20 = copy.deepcopy(create)
create_v20['properties']['flavor']['properties']['name'] = (
    parameter_types.name_with_leading_trailing_spaces
)


# 2.55 adds an optional description field with a max length of 65535 since the
# backing database column is a TEXT column which is 64KiB.
_flavor_description = {
    'type': ['string', 'null'], 'minLength': 0, 'maxLength': 65535,
    'pattern': parameter_types.valid_description_regex,
}


create_v255 = copy.deepcopy(create)
create_v255['properties']['flavor']['properties']['description'] = (
    _flavor_description)


update = {
    'type': 'object',
    'properties': {
        'flavor': {
            'type': 'object',
            'properties': {
                'description': _flavor_description
            },
            # Since the only property that can be specified on update is the
            # description field, it is required. If we allow updating other
            # flavor attributes in a later microversion, we should reconsider
            # what is required.
            'required': ['description'],
            'additionalProperties': False,
        },
    },
    'required': ['flavor'],
    'additionalProperties': False,
}

index_query = {
    'type': 'object',
    'properties': {
        'limit': parameter_types.multi_params(
             parameter_types.non_negative_integer),
        'marker': parameter_types.multi_params({'type': 'string'}),
        'is_public': parameter_types.multi_params({'type': 'string'}),
        'minRam': parameter_types.multi_params({'type': 'string'}),
        'minDisk': parameter_types.multi_params({'type': 'string'}),
        'sort_key': parameter_types.multi_params({'type': 'string',
                                                  'enum': VALID_SORT_KEYS}),
        'sort_dir': parameter_types.multi_params({'type': 'string',
                                                  'enum': VALID_SORT_DIR})
    },
    # NOTE(gmann): This is kept True to keep backward compatibility.
    # As of now Schema validation stripped out the additional parameters and
    # does not raise 400. In microversion 2.75, we have blocked the additional
    # parameters.
    'additionalProperties': True
}

index_query_275 = copy.deepcopy(index_query)
index_query_275['additionalProperties'] = False

# TODO(stephenfin): Remove additionalProperties in a future API version
show_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

_flavor_basic = {
    'type': 'object',
    'properties': {
        'id': {'type': 'string'},
        'links': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'href': {'type': 'string', 'format': 'uri'},
                    'rel': {'type': 'string'},
                },
                'required': ['href', 'rel'],
                'additionalProperties': False,
            },
        },
        'name': {'type': 'string'},
    },
    'required': ['id', 'links', 'name'],
    'additionalProperties': False,
}

_flavor_basic_v255 = copy.deepcopy(_flavor_basic)
_flavor_basic_v255['properties']['description'] = {'type': ['string', 'null']}
_flavor_basic_v255['required'].append('description')

_flavor = {
    'type': 'object',
    'properties': {
        'disk': {'type': 'integer'},
        'id': {'type': 'string'},
        'links': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'href': {'type': 'string', 'format': 'uri'},
                    'rel': {'type': 'string'},
                },
                'required': ['href', 'rel'],
            },
        },
        'name': {'type': 'string'},
        'os-flavor-access:is_public': {},
        'ram': {'type': 'integer'},
        'rxtx_factor': {},
        'swap': {
            'anyOf': [
                {'type': 'integer'},
                {'const': ''},
            ],
        },
        'vcpus': {'type': 'integer'},
        'OS-FLV-EXT-DATA:ephemeral': {'type': 'integer'},
        'OS-FLV-DISABLED:disabled': {'type': 'boolean'},
    },
    'required': [
        'disk',
        'id',
        'links',
        'name',
        'os-flavor-access:is_public',
        'ram',
        'rxtx_factor',
        'swap',
        'vcpus',
        'OS-FLV-DISABLED:disabled',
        'OS-FLV-EXT-DATA:ephemeral',
    ],
    'additionalProperties': False,
}

_flavor_v255 = copy.deepcopy(_flavor)
_flavor_v255['properties']['description'] = {'type': ['string', 'null']}
_flavor_v255['required'].append('description')

_flavor_v261 = copy.deepcopy(_flavor_v255)
_flavor_v261['properties']['extra_specs'] = {
    'type': 'object',
    'patternProperties': {
        '^[a-zA-Z0-9-_:. ]{1,255}$': {'type': 'string', 'maxLength': 255},
    },
    'additionalProperties': False,
}

_flavor_v275 = copy.deepcopy(_flavor_v261)
# we completely overwrite this since the new variant is much simpler
_flavor_v275['properties']['swap'] = {'type': 'integer'}

_flavors_links = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'href': {'type': 'string', 'format': 'uri'},
            'rel': {'type': 'string'},
        },
        'required': ['href', 'rel'],
        'additionalProperties': False,
    },
}

delete_response = {
    'type': 'null',
}

create_response = {
    'type': 'object',
    'properties': {
        'flavor': copy.deepcopy(_flavor),
    },
    'required': ['flavor'],
    'additionalProperties': False,
}

create_response_v255 = copy.deepcopy(create_response)
create_response_v255['properties']['flavor'] = copy.deepcopy(_flavor_v255)

create_response_v261 = copy.deepcopy(create_response_v255)
create_response_v261['properties']['flavor'] = copy.deepcopy(_flavor_v261)

create_response_v275 = copy.deepcopy(create_response_v261)
create_response_v275['properties']['flavor'] = copy.deepcopy(_flavor_v275)

# NOTE(stephenfin): update is only available from 2.55 and the response is
# identical to the create and show response from that point forward
update_response = {
    'type': 'object',
    'properties': {
        'flavor': copy.deepcopy(_flavor_v255),
    },
    'required': ['flavor'],
    'additionalProperties': False,
}

update_response_v261 = copy.deepcopy(update_response)
update_response_v261['properties']['flavor'] = copy.deepcopy(_flavor_v261)

update_response_v275 = copy.deepcopy(update_response_v261)
update_response_v275['properties']['flavor'] = copy.deepcopy(_flavor_v275)

index_response = {
    'type': 'object',
    'properties': {
        'flavors': {
            'type': 'array',
            'items': _flavor_basic,
        },
        'flavors_links': _flavors_links,
    },
    'required': ['flavors'],
    'additionalProperties': False,
}

index_response_v255 = copy.deepcopy(index_response)
index_response_v255['properties']['flavors']['items'] = _flavor_basic_v255

detail_response = {
    'type': 'object',
    'properties': {
        'flavors': {
            'type': 'array',
            'items': _flavor,
        },
        'flavors_links': _flavors_links,
    },
    'required': ['flavors'],
    'additionalProperties': False,
}

detail_response_v255 = copy.deepcopy(detail_response)
detail_response_v255['properties']['flavors']['items'] = _flavor_v255

detail_response_v261 = copy.deepcopy(detail_response_v255)
detail_response_v261['properties']['flavors']['items'] = _flavor_v261

detail_response_v275 = copy.deepcopy(detail_response_v261)
detail_response_v275['properties']['flavors']['items'] = _flavor_v275

show_response = {
    'type': 'object',
    'properties': {
        'flavor': copy.deepcopy(_flavor),
    },
    'required': ['flavor'],
    'additionalProperties': False,
}

show_response_v255 = copy.deepcopy(show_response)
show_response_v255['properties']['flavor'] = copy.deepcopy(_flavor_v255)

show_response_v261 = copy.deepcopy(show_response_v255)
show_response_v261['properties']['flavor'] = copy.deepcopy(_flavor_v261)

show_response_v275 = copy.deepcopy(show_response_v261)
show_response_v275['properties']['flavor'] = copy.deepcopy(_flavor_v275)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add_constraint_instance_share_avoid_duplicates

Revision ID: d60bddf7a903
Revises: 13863f4e1612
Create Date: 2024-03-06 17:05:29.361678
"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd60bddf7a903'
down_revision = '13863f4e1612'
branch_labels = None
depends_on = None


def upgrade():
    op.drop_table("share_mapping")
    op.create_table(
        "share_mapping",
        sa.Column("created_at", sa.DateTime),
        sa.Column("updated_at", sa.DateTime),
        sa.Column(
            "id",
            sa.BigInteger().with_variant(sa.Integer, "sqlite"),
            primary_key=True,
            autoincrement=True,
            nullable=False,
        ),
        sa.Column("uuid", sa.String(36)),
        sa.Column(
            "instance_uuid",
            sa.String(length=36),
            sa.ForeignKey(
                "instances.uuid", name="share_mapping_instance_uuid_fkey"
            ),
        ),
        sa.Column("share_id", sa.String(length=36)),
        sa.Column("status", sa.String(length=32)),
        sa.Column("tag", sa.String(48)),
        sa.Column("export_location", sa.Text),
        sa.Column("share_proto", sa.String(32)),
        sa.UniqueConstraint(
            "instance_uuid",
            "share_id",
            name="uniq_key_pairs0instance_uuid0share_id",
        ),
        sa.UniqueConstraint(
            "instance_uuid",
            "tag",
            name="uniq_key_pairs0instance_uuid0tag",
        ),
        sa.Index("share_idx", "share_id"),
        sa.Index(
            "share_mapping_instance_uuid_share_id_idx",
            "instance_uuid",
            "share_id",
        ),
        mysql_engine="InnoDB",
        mysql_charset="utf8",
    )



# Copyright 2016 Hewlett Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_middleware import cors


def set_defaults():
    """Update default configuration options for oslo.middleware."""
    cors.set_defaults(
        allow_headers=['X-Auth-Token',
                       'X-Openstack-Request-Id',
                       'X-Identity-Status',
                       'X-Roles',
                       'X-Service-Catalog',
                       'X-User-Id',
                       'X-Tenant-Id',
                       'X-OpenStack-Nova-API-Version',
                       'OpenStack-API-Version'],
        expose_headers=['X-Auth-Token',
                        'X-Openstack-Request-Id',
                        'X-Subject-Token',
                        'X-Service-Token',
                        'X-OpenStack-Nova-API-Version',
                        'OpenStack-API-Version'],
        allow_methods=['GET',
                       'PUT',
                       'POST',
                       'DELETE',
                       'PATCH']
    )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova.compute import api as compute
from nova import exception
from nova.tests import fixtures as nova_fixtures
from nova.tests.functional.api import client
from nova.tests.functional.api_sample_tests import test_servers
from oslo_utils.fixture import uuidsentinel
from unittest import mock


class ServerSharesBase(test_servers.ServersSampleBase):
    sample_dir = 'os-server-shares'
    microversion = '2.97'
    scenarios = [('v2_97', {'api_major_version': 'v2.1'})]

    def setUp(self):
        super(ServerSharesBase, self).setUp()
        self.manila_fixture = self.useFixture(nova_fixtures.ManilaFixture())
        self.compute_api = compute.API()

    def _get_create_subs(self):
        return {'share_id': 'e8debdc0-447a-4376-a10a-4cd9122d7986',
                'uuid': '[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}'
                '-[0-9a-f]{4}-[0-9a-f]{12}',
                }

    def create_server_ok(self, requested_flavor=None):
        flavor = self._create_flavor(extra_spec=requested_flavor)
        server = self._create_server(networks='auto', flavor_id=flavor)
        self._stop_server(server)
        return server['id']

    def create_server_not_stopped(self):
        server = self._create_server(networks='auto')
        return server['id']

    def _post_server_shares(self):
        """Verify the response status and returns the UUID of the
        newly created server with shares.
        """
        uuid = self.create_server_ok()
        subs = self._get_create_subs()
        response = self._do_post(
            "servers/%s/shares" % uuid, "server-shares-create-req", subs
        )

        self._verify_response(
            'server-shares-create-resp', subs, response, 201)

        return uuid


class ServerSharesJsonTest(ServerSharesBase):
    def test_server_shares_create(self):
        """Verify we can create a share mapping.
        """
        self._post_server_shares()

    def test_server_shares_create_fails_if_already_created(self):
        """Verify we cannot create a share mapping already created.
        """
        uuid = self._post_server_shares()
        # Following mock simulate a race condition between 2 requests that
        # would hit the share_mapping.create() almost at the same time.
        with mock.patch(
            "nova.db.main.api.share_mapping_get_by_instance_uuid_and_share_id"
        ) as mock_db:
            mock_db.return_value = None
            subs = self._get_create_subs()
            response = self._do_post(
                "servers/%s/shares" % uuid, "server-shares-create-req", subs
            )
            self.assertEqual(409, response.status_code)
            self.assertIn('already associated to this server', response.text)

    def test_server_shares_create_with_tag_fails_if_already_created(self):
        """Verify we cannot create a share mapping with a new tag if it is
        already created.
        """
        uuid = self._post_server_shares()
        subs = self._get_create_subs()
        subs['tag'] = "my-tag"
        response = self._do_post(
            "servers/%s/shares" % uuid, "server-shares-create-tag-req", subs
        )
        self.assertEqual(409, response.status_code)
        self.assertIn(
            "Share 'e8debdc0-447a-4376-a10a-4cd9122d7986' or "
            "tag 'my-tag' already associated to this server.",
            response.text,
        )

    def test_server_shares_create_fails_instance_not_stopped(self):
        """Verify we cannot create a share if instance is not stopped.
        """
        uuid = self.create_server_not_stopped()
        subs = self._get_create_subs()
        response = self._do_post(
            "servers/%s/shares" % uuid, "server-shares-create-req", subs
        )
        self.assertEqual(409, response.status_code)
        self.assertIn('while it is in vm_state active', response.text)

    def test_server_shares_create_fails_incorrect_configuration(self):
        """Verify we cannot create a share we don't have the
        appropriate configuration.
        """
        with mock.patch.dict(self.compute.driver.capabilities,
                             supports_mem_backing_file=False):
            self.compute.stop()
            self.compute.start()
            uuid = self.create_server_ok()
            subs = self._get_create_subs()
            response = self._do_post('servers/%s/shares' % uuid,
                                    'server-shares-create-req', subs)
            self.assertEqual(409, response.status_code)
            self.assertIn(
                'Feature not supported because either compute or '
                'instance are not configured correctly.', response.text
            )

    def test_server_shares_create_fails_cannot_allow_policy(self):
        """Verify we raise an exception if we get a timeout to apply policy"""
        uuid = self.create_server_ok()
        subs = self._get_create_subs()
        # simulate that manila does not set the requested access in time and
        # nova times out waiting for it.
        self.manila_fixture.mock_get_access.return_value = None
        self.manila_fixture.mock_get_access.side_effect = None
        self.flags(share_apply_policy_timeout=2, group='manila')

        # Here we are using CastAsCallFixture so we got an exception from
        # nova api. This should not happen without the fixture.
        response = self._do_post(
            "servers/%s/shares" % uuid, "server-shares-create-req", subs
        )
        self.assertEqual(500, response.status_code)
        self.assertIn(
            "nova.exception.ShareAccessGrantError",
            response.text,
        )

    def test_server_shares_create_with_alternative_flavor(self):
        """Verify we can create a share with the proper flavor.
        """
        with mock.patch.dict(self.compute.driver.capabilities,
                             supports_mem_backing_file=False):
            self.compute.stop()
            self.compute.start()
            uuid = self.create_server_ok(
                requested_flavor={"hw:mem_page_size": "large"}
            )
            subs = self._get_create_subs()
            response = self._do_post(
                "servers/%s/shares" % uuid, "server-shares-create-req", subs
            )
            self.assertEqual(201, response.status_code)

    def test_server_shares_create_fails_share_not_found(self):
        """Verify we can not create a share if the share does not
        exists.
        """
        self.manila_fixture.mock_get.side_effect = exception.ShareNotFound(
            share_id='fake_uuid')
        uuid = self.create_server_ok()
        subs = self._get_create_subs()
        response = self._do_post(
            "servers/%s/shares" % uuid, "server-shares-create-req", subs
        )
        self.assertEqual(404, response.status_code)
        self.assertIn("Share fake_uuid could not be found", response.text)

    def test_server_shares_create_unknown_instance(self):
        """Verify creating a share on an unknown instance reports an error.
        """
        self.create_server_ok()
        subs = self._get_create_subs()
        response = self._do_post(
            "servers/%s/shares" % uuidsentinel.fake_uuid,
            "server-shares-create-req",
            subs,
        )
        self.assertEqual(404, response.status_code)
        self.assertIn("could not be found", response.text)

    def test_server_shares_create_fails_share_in_error(self):
        """Verify creating a share which is in error reports an error.
        """
        uuid = self.create_server_ok()
        subs = self._get_create_subs()
        self.manila_fixture.mock_get.side_effect = (
            self.manila_fixture.fake_get_share_status_error
        )

        response = self._do_post(
            "servers/%s/shares" % uuid, "server-shares-create-req", subs
        )
        self.assertEqual(409, response.status_code)
        self.assertIn(
            "Share e8debdc0-447a-4376-a10a-4cd9122d7986 is in 'error' "
            "instead of 'available' status.",
            response.text,
        )

    def test_server_shares_create_fails_export_location_missing(self):
        """Verify creating a share without export location reports an error.
        """
        uuid = self.create_server_ok()
        subs = self._get_create_subs()
        self.manila_fixture.mock_get.side_effect = (
            self.manila_fixture.fake_get_share_export_location_missing
        )

        response = self._do_post(
            "servers/%s/shares" % uuid, "server-shares-create-req", subs
        )
        self.assertEqual(409, response.status_code)
        self.assertIn(
            "Share e8debdc0-447a-4376-a10a-4cd9122d7986 export location is "
            "missing.",
            response.text,
        )

    def test_server_shares_create_fails_unknown_protocol(self):
        """Verify creating a share with an unknown protocol reports an error.
        """
        uuid = self.create_server_ok()
        subs = self._get_create_subs()
        self.manila_fixture.mock_get.side_effect = (
            self.manila_fixture.fake_get_share_unknown_protocol
        )

        response = self._do_post(
            "servers/%s/shares" % uuid, "server-shares-create-req", subs
        )
        self.assertEqual(409, response.status_code)
        self.assertIn("Share protocol CIFS is not supported.", response.text)

    def test_server_shares_index(self):
        """Verify we can list shares.
        """
        uuid = self._post_server_shares()
        subs = self._get_create_subs()
        response = self._do_get("servers/%s/shares" % uuid)
        self._verify_response("server-shares-list-resp", subs, response, 200)

    def test_server_shares_index_unknown_instance(self):
        """Verify getting shares on an unknown instance reports an error.
        """
        response = self._do_get('servers/%s/shares' % uuidsentinel.fake_uuid)
        self.assertEqual(404, response.status_code)
        self.assertIn(
            "could not be found",
            response.text
        )

    def test_server_shares_show(self):
        """Verify we can show a share.
        """
        uuid = self._post_server_shares()
        subs = self._get_create_subs()
        response = self._do_get(
            "servers/%s/shares/%s" % (uuid, subs["share_id"])
        )
        self._verify_response("server-shares-show-resp", subs, response, 200)

    def test_server_shares_show_fails_share_not_found(self):
        """Verify we can not show a share if the share does not
        exists.
        """
        uuid = self.create_server_ok()
        subs = self._get_create_subs()
        response = self._do_get(
            "servers/%s/shares/%s" % (uuid, subs["share_id"])
        )
        self.assertEqual(404, response.status_code)
        self.assertIn(
            "Share e8debdc0-447a-4376-a10a-4cd9122d7986 could not be found",
            response.text
        )

    def test_server_shares_show_unknown_instance(self):
        """Verify showing a share on an unknown instance reports an error.
        """
        self._post_server_shares()
        subs = self._get_create_subs()
        response = self._do_get(
            "servers/%s/shares/%s" % (uuidsentinel.fake_uuid, subs["share_id"])
        )
        self.assertEqual(404, response.status_code)
        self.assertIn(
            "could not be found",
            response.text
        )

    def test_server_shares_delete(self):
        """Verify we can delete share.
        """
        uuid = self._post_server_shares()
        subs = self._get_create_subs()
        response = self._do_delete(
            "servers/%s/shares/%s" % (uuid, subs["share_id"])
        )
        self.assertEqual(200, response.status_code)

        # Check share is not anymore available
        response = self._do_get(
            "servers/%s/shares/%s" % (uuid, subs["share_id"])
        )
        self.assertEqual(404, response.status_code)

    def test_server_shares_delete_instance(self):
        """Verify we can delete an instance and its associated share is
        deleted as well.
        """
        uuid = self._post_server_shares()
        subs = self._get_create_subs()

        # Check share is created
        response = self._do_get(
            "servers/%s/shares/%s" % (uuid, subs["share_id"])
        )
        self._verify_response("server-shares-show-resp", subs, response, 200)

        # Delete the instance
        response = self._do_delete(
            "servers/%s" % (uuid)
        )
        self.assertEqual(204, response.status_code)

        # Check share is not anymore available
        response = self._do_get(
            "servers/%s/shares/%s" % (uuid, subs["share_id"])
        )
        self.assertEqual(404, response.status_code)

    def test_server_shares_delete_fails_share_not_found(self):
        """Verify we have an error if we want to remove an unknown share.
        """
        uuid = self._post_server_shares()
        response = self._do_delete(
            "servers/%s/shares/%s" % (uuid, uuidsentinel.wrong_share_id)
        )
        self.assertEqual(404, response.status_code)

    def test_server_shares_delete_fails_instance_not_stopped(self):
        """Verify we cannot remove a share if the instance is not stopped.
        """
        uuid = self._post_server()
        subs = self._get_create_subs()
        response = self._do_post(
            "servers/%s/shares" % uuid, "server-shares-delete-req", subs
        )
        response = self._do_delete(
            "servers/%s/shares/%s" % (uuid, subs["share_id"])
        )
        self.assertEqual(409, response.status_code)

    def test_server_shares_delete_unknown_instance(self):
        """Verify deleting a share on an unknown instance reports an error.
        """
        uuid = self._post_server_shares()
        subs = self._get_create_subs()
        response = self._do_post(
            "servers/%s/shares" % uuid, "server-shares-delete-req", subs
        )
        response = self._do_delete(
            "servers/%s/shares/%s" % (uuidsentinel.fake_uuid, subs["share_id"])
        )
        self.assertEqual(404, response.status_code)
        self.assertIn(
            "could not be found",
            response.text
        )

    def test_server_shares_delete_fails_cannot_deny_policy(self):
        """Verify we raise an exception if we cannot deny the policy.
        """
        uuid = self._post_server_shares()
        subs = self._get_create_subs()
        self.manila_fixture.mock_deny.return_value = None
        self.manila_fixture.mock_deny.side_effect = (
            exception.ShareAccessRemovalError(
                share_id=subs["share_id"],
                reason="Resource could not be found.",
            )
        )

        # Here we are using CastAsCallFixture so we got an exception from
        # nova api. This should not happen without the fixture.
        response = self._do_delete(
            "servers/%s/shares/%s" % (uuid, subs["share_id"])
        )
        self.assertEqual(500, response.status_code)
        self.assertIn('nova.exception.ShareAccessRemovalError', response.text)


class ServerSharesJsonAdminTest(ServerSharesBase):
    ADMIN_API = True

    def _post_server_shares(self):
        """Verify the response status and returns the UUID of the
        newly created server with shares.
        """
        uuid = self.create_server_ok()
        subs = self._get_create_subs()
        response = self._do_post(
            "servers/%s/shares" % uuid, "server-shares-create-req", subs
        )
        self._verify_response(
            'server-shares-admin-create-resp', subs, response, 201)

        return uuid

    def test_server_shares_create(self):
        """Verify we can create a share mapping.
        """
        self._post_server_shares()

    def test_server_shares_show(self):
        """Verify we can show a share as admin and thus have more
           information.
        """
        uuid = self._post_server_shares()
        subs = self._get_create_subs()
        response = self._do_get(
            "servers/%s/shares/%s" % (uuid, subs["share_id"])
        )
        self._verify_response(
            "server-shares-admin-show-resp", subs, response, 200
        )

    def _block_action(self, body):
        uuid = self._post_server_shares()

        ex = self.assertRaises(
            client.OpenStackApiException,
            self.api.post_server_action,
            uuid,
            body
        )

        self.assertEqual(409, ex.response.status_code)
        self.assertIn(
            "Feature not supported with instances that have shares.",
            ex.response.text
        )

    def test_shelve_server_with_share_fails(self):
        self._block_action({"shelve": None})

    def test_suspend_server_with_share_fails(self):
        self._block_action({"suspend": None})

    def test_evacuate_server_with_share_fails(self):
        self._block_action({"evacuate": {}})

    def test_resize_server_with_share_fails(self):
        self._block_action({"resize": {"flavorRef": "2"}})

    def test_migrate_server_with_share_fails(self):
        self._block_action({"migrate": None})

    def test_live_migrate_server_with_share_fails(self):
        self._block_action(
            {"os-migrateLive": {
                "host": None,
                "block_migration": "auto"
                }
             }
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import jsonschema.exceptions

from nova.api.openstack import compute
from nova.api.validation import validators
from nova import test


class SchemaTest(test.NoDBTestCase):

    def setUp(self):
        super().setUp()
        self.router = compute.APIRouterV21()
        self.meta_schema = validators._SchemaValidator.validator_org

    def test_schemas(self):
        missing_request_schemas = set()
        missing_query_schemas = set()
        invalid_schemas = set()

        def _validate_func(func, method):
            if method in ("POST", "PUT", "PATCH"):
                # request body validation
                if not hasattr(func, '_request_schema'):
                    missing_request_schemas.add(func.__qualname__)
                else:
                    try:
                        self.meta_schema.check_schema(func._request_schema)
                    except jsonschema.exceptions.SchemaError:
                        invalid_schemas.add(func.__qualname__)
            elif method in ("GET",):
                # request query string validation
                if not hasattr(func, '_query_schema'):
                    missing_query_schemas.add(func.__qualname__)
                else:
                    try:
                        self.meta_schema.check_schema(func._query_schema)
                    except jsonschema.exceptions.SchemaError:
                        invalid_schemas.add(func.__qualname__)

            # TODO(stephenfin): Check for missing schemas once we have added
            # them all
            if hasattr(func, '_response_schema'):
                try:
                    self.meta_schema.check_schema(func._response_schema)
                except jsonschema.exceptions.SchemaError:
                    invalid_schemas.add(func.__qualname__)

        for route in self.router.map.matchlist:
            if 'controller' not in route.defaults:
                continue

            controller = route.defaults['controller']

            path = ""
            for part in route.routelist:
                if isinstance(part, dict):
                    path += "{" + part["name"] + "}"
                else:
                    path += part

            method = (
                route.conditions.get("method", "GET")[0]
                if route.conditions
                else "GET"
            )
            action = route.defaults["action"]

            if path.endswith('/action'):
                # all actions should use POST
                assert method == 'POST'

                wsgi_actions = [
                    (k, v, controller.controller) for k, v in
                    controller.controller.wsgi_actions.items()
                ]
                for sub_controller in controller.sub_controllers:
                    wsgi_actions += [
                        (k, v, sub_controller) for k, v in
                        sub_controller.wsgi_actions.items()
                    ]

                for (
                    wsgi_action, wsgi_method, action_controller
                ) in wsgi_actions:
                    func = controller.wsgi_actions[wsgi_action]

                    if hasattr(action_controller, 'versioned_methods'):
                        if wsgi_method in action_controller.versioned_methods:
                            # currently all our actions are unversioned and if
                            # this changes then we need to fix this
                            funcs = action_controller.versioned_methods[
                                wsgi_method
                            ]
                            assert len(funcs) == 1
                            func = funcs[0].func

                    # method will always be POST for actions
                    _validate_func(func, method)
            else:
                # body validation
                versioned_methods = getattr(
                    controller.controller, 'versioned_methods', {}
                )
                if action in versioned_methods:
                    # versioned method
                    for versioned_method in sorted(
                        versioned_methods[action],
                        key=lambda v: v.start_version
                    ):
                        func = versioned_method.func

                        _validate_func(func, method)
                else:
                    # unversioned method
                    func = getattr(controller.controller, action)
                    _validate_func(func, method)

        if missing_request_schemas:
            raise test.TestingException(
                f"Found API resources without schemas: "
                f"{sorted(missing_request_schemas)}"
            )

        if missing_query_schemas:
            raise test.TestingException(
                f"Found API resources without query schemas: "
                f"{sorted(missing_query_schemas)}"
            )

        if invalid_schemas:
            raise test.TestingException(
                f"Found API resources with invalid schemas: "
                f"{sorted(invalid_schemas)}"
            )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from requests import Response

import fixtures
from keystoneauth1 import loading as ks_loading

import nova.conf
from nova import context as nova_context
from nova import exception
from nova.share import manila
from nova import test

from openstack import exceptions as sdk_exc
from openstack.shared_file_system.v2 import (
    resource_locks as sdk_resource_locks
)

from openstack.shared_file_system.v2 import (
    share_access_rule as sdk_share_access_rule
)
from openstack.shared_file_system.v2 import (
    share_export_locations as sdk_share_export_locations
)
from openstack.shared_file_system.v2 import share as sdk_share

from openstack import utils
from unittest import mock

from nova.tests.unit.api.openstack import fakes

CONF = nova.conf.CONF


def stub_share(share_id):
    share = sdk_share.Share()
    share.id = share_id
    share.size = 1
    share.availability_zone = "nova"
    share.created_at = "2015-09-18T10:25:24.000000"
    share.status = "available"
    share.name = "share_London"
    share.description = "My custom share London"
    share.project_id = "16e1ab15c35a457e9c2b2aa189f544e1"
    share.snapshot_id = None
    share.share_network_id = "713df749-aac0-4a54-af52-10f6c991e80c"
    share.share_protocol = "NFS"
    share.metadata = {
            "project": "my_app",
            "aim": "doc"
        }
    share.share_type = "25747776-08e5-494f-ab40-a64b9d20d8f7"
    share.is_public = True
    share.share_server_id = "e268f4aa-d571-43dd-9ab3-f49ad06ffaef"
    share.host = "manila2@generic1#GENERIC1"

    share.location = utils.Munch(
        {
            "cloud": "envvars",
            "region_name": "RegionOne",
            "zone": "manila-zone-0",
            "project": utils.Munch(
                {
                    "id": "bce4fcc3bd0d4c598f610cb45ec5c5ba",
                    "name": "demo",
                    "domain_id": "default",
                    "domain_name": None,
                }
            ),
        }
    )
    return share


def stub_export_locations():
    export_locations = []
    export_location = sdk_share_export_locations.ShareExportLocation()
    export_location.id = "b6bd76ce-12a2-42a9-a30a-8a43b503867d"
    export_location.path = (
        "10.0.0.3:/shares/share-e1c2d35e-fe67-4028-ad7a-45f668732b1d"
    )
    export_location.is_preferred = True
    export_location.share_instance_id = (
        "e1c2d35e-fe67-4028-ad7a-45f668732b1d"
    )
    export_location.is_admin = True
    export_location.share_instance_id = "e1c2d35e-fe67-4028-ad7a-45f668732b1d"
    export_location.location = utils.Munch(
        {
            "cloud": "envvars",
            "region_name": "RegionOne",
            "zone": None,
            "project": utils.Munch(
                {
                    "id": "bce4fcc3bd0d4c598f610cb45ec5c5ba",
                    "name": "demo",
                    "domain_id": "default",
                    "domain_name": None,
                }
            ),
        }
    )

    export_locations.append(export_location)
    for item in export_locations:
        yield item


def stub_access_list():
    access_list = []
    access_list.append(stub_access())
    for access in access_list:
        yield access


def stub_access():
    access = sdk_share_access_rule.ShareAccessRule()
    access.id = "a25b2df3-90bd-4add-afa6-5f0dbbd50452"
    access.access_level = "rw"
    access.access_to = "0.0.0.0/0"
    access.access_type = "ip"
    access.state = "active"
    access.access_key = None
    access.created_at = "2023-07-21T15:20:01.812350"
    access.updated_at = "2023-07-21T15:20:01.812350"
    access.metadata = {}
    access.location = utils.Munch(
        {
            "cloud": "envvars",
            "region_name": "RegionOne",
            "zone": None,
            "project": utils.Munch(
                {
                    "id": "bce4fcc3bd0d4c598f610cb45ec5c5ba",
                    "name": "demo",
                    "domain_id": "default",
                    "domain_name": None,
                }
            ),
        }
    )
    return access


def stub_lock(share_id):
    lock = sdk_resource_locks.ResourceLock()
    lock.id = "a37b7da7-5d72-49d3-bf3b-aebd64828089"
    lock.project_id = "ded249b25f6f46918fef4e69f427590c"
    lock.resource_type = "share"
    lock.resource_id = share_id
    lock.resource_action = "delete"
    lock.lock_reason = "nova lock"
    lock.created_at = "2023-07-31T09:39:38.441320"
    lock.updated_at = None
    lock.location = utils.Munch(
        {
            "cloud": "envvars",
            "region_name": "RegionOne",
            "zone": None,
            "project": utils.Munch(
                {
                    "id": "bce4fcc3bd0d4c598f610cb45ec5c5ba",
                    "name": "demo",
                    "domain_id": "default",
                    "domain_name": None,
                }
            ),
        }
    )
    return lock


def stub_resource_locks(share_id):
    resource_locks = []
    resource_lock = stub_lock(share_id)
    resource_locks.append(resource_lock)
    for lock in resource_locks:
        yield lock


class BaseManilaTestCase(object):
    project_id = fakes.FAKE_PROJECT_ID

    def setUp(self):

        super(BaseManilaTestCase, self).setUp()

        self.mock_get_confgrp = self.useFixture(fixtures.MockPatch(
            'nova.utils._get_conf_group')).mock

        self.mock_ks_loading = self.useFixture(
            fixtures.MockPatchObject(ks_loading, 'load_auth_from_conf_options')
        ).mock

        self.service_type = 'shared-file-system'
        self.mock_connection = self.useFixture(
            fixtures.MockPatch(
                "nova.utils.connection.Connection", side_effect=self.fake_conn
            )
        ).mock

        # We need to stub the CONF global in nova.utils to assert that the
        # Connection constructor picks it up.
        self.mock_conf = self.useFixture(fixtures.MockPatch(
            'nova.utils.CONF')).mock

        self.api = manila.API()

        self.context = nova_context.RequestContext(
            user_id="fake_user", project_id=self.project_id
        )

    def fake_conn(self, *args, **kwargs):
        class FakeConnection(object):
            def __init__(self):
                self.shared_file_system = FakeConnectionShareV2Proxy()

        class FakeConnectionShareV2Proxy(object):
            def __init__(self):
                pass

            def get_share(self, share_id):
                if share_id == 'nonexisting':
                    raise sdk_exc.ResourceNotFound
                return stub_share(share_id)

            def export_locations(self, share_id):
                return stub_export_locations()

            def access_rules(self, share_id):
                if share_id == 'nonexisting':
                    raise sdk_exc.ResourceNotFound
                if share_id == 'nonexisting2':
                    raise sdk_exc.ResourceNotFound
                if share_id == '4567':
                    return []
                return stub_access_list()

            def create_access_rule(self, share_id, **kwargs):
                if share_id == '2345':
                    raise sdk_exc.BadRequestException
                if share_id == 'nonexisting':
                    raise sdk_exc.ResourceNotFound
                return stub_access()

            def delete_access_rule(self, access_id, share_id, unrestrict):
                if share_id == 'nonexisting':
                    raise sdk_exc.ResourceNotFound
                res = Response()
                res.status_code = 202
                res.reason = "Internal error"
                if share_id == '2345':
                    res.status_code = 500
                return res

            def get_all_resource_locks(self, resource_id=None):
                if resource_id == '1234':
                    return []
                if resource_id == '2345':
                    return []
                if resource_id == 'nonexisting':
                    return []
                return stub_resource_locks(resource_id)

            def create_resource_lock(
                self, resource_id=None, resource_type=None, lock_reason=None
            ):
                if resource_id == "nonexisting":
                    raise sdk_exc.BadRequestException
                return stub_lock(resource_id)

            def delete_resource_lock(self, share_id):
                pass

        return FakeConnection()

    @mock.patch('nova.utils.get_sdk_adapter')
    def test_client(self, mock_get_sdk_adapter):
        client = manila._manilaclient(self.context)
        self.assertTrue(hasattr(client, 'get_share'))
        self.assertTrue(hasattr(client, 'export_locations'))
        self.assertTrue(hasattr(client, 'access_rules'))
        self.assertTrue(hasattr(client, 'create_access_rule'))
        self.assertTrue(hasattr(client, 'delete_access_rule'))
        mock_get_sdk_adapter.assert_called_once_with(
            "shared-file-system",
            admin=False,
            check_service=True,
            context=self.context,
            shared_file_system_api_version="2.82",
            global_request_id=self.context.global_id,
        )

    @mock.patch('nova.utils.get_sdk_adapter')
    def test_client_admin(self, mock_get_sdk_adapter):
        client = manila._manilaclient(self.context, admin=True)
        self.assertTrue(hasattr(client, 'get_share'))
        self.assertTrue(hasattr(client, 'export_locations'))
        self.assertTrue(hasattr(client, 'access_rules'))
        self.assertTrue(hasattr(client, 'create_access_rule'))
        self.assertTrue(hasattr(client, 'delete_access_rule'))
        mock_get_sdk_adapter.assert_called_once_with(
            "shared-file-system",
            admin=True,
            check_service=True,
            context=self.context,
            shared_file_system_api_version="2.82",
            global_request_id=self.context.global_id,
        )


class ManilaTestCase(BaseManilaTestCase, test.NoDBTestCase):
    def test_get_fails_non_existing_share(self):
        """Tests that we fail if trying to get an
        non existing share.
        """
        exc = self.assertRaises(
            exception.ShareNotFound, self.api.get, self.context, "nonexisting"
        )

        self.assertIn("Share nonexisting could not be found.", exc.message)

    @mock.patch(
        'nova.utils.get_sdk_adapter', side_effect=nova.utils.get_sdk_adapter)
    def test_get_share(self, mock_get_sdk_adapter):
        """Tests that we manage to get a share.
        """
        share = self.api.get(self.context, '1234')
        mock_get_sdk_adapter.assert_called_once_with(
            "shared-file-system",
            admin=False,
            check_service=True,
            context=self.context,
            shared_file_system_api_version="2.82",
            global_request_id=self.context.global_id,
        )
        self.assertIsInstance(share, manila.Share)
        self.assertEqual('1234', share.id)
        self.assertEqual(1, share.size)
        self.assertEqual('nova', share.availability_zone)
        self.assertEqual('2015-09-18T10:25:24.000000',
                         share.created_at)
        self.assertEqual('available', share.status)
        self.assertEqual('share_London', share.name)
        self.assertEqual('My custom share London',
                         share.description)
        self.assertEqual('16e1ab15c35a457e9c2b2aa189f544e1',
                         share.project_id)
        self.assertIsNone(share.snapshot_id)
        self.assertEqual(
            '713df749-aac0-4a54-af52-10f6c991e80c',
            share.share_network_id)
        self.assertEqual('NFS', share.share_proto)
        self.assertEqual(share.export_location,
                "10.0.0.3:/shares/"
                "share-e1c2d35e-fe67-4028-ad7a-45f668732b1d"
                )
        self.assertEqual({"project": "my_app", "aim": "doc"},
                         share.metadata)
        self.assertEqual(
            '25747776-08e5-494f-ab40-a64b9d20d8f7',
            share.share_type)
        self.assertTrue(share.is_public)

    def test_get_access_fails_non_existing_share(self):
        """Tests that we fail if trying to get an access on a
        non existing share.
        """
        exc = self.assertRaises(
            exception.ShareNotFound,
            self.api.get_access,
            self.context,
            "nonexisting",
            "ip",
            "0.0.0.0/0",
        )

        self.assertIn("Share nonexisting could not be found.", exc.message)

        exc = self.assertRaises(
            exception.ShareNotFound,
            self.api.get_access,
            self.context,
            "nonexisting2",
            "ip",
            "0.0.0.0/0",
        )

        self.assertIn("Share nonexisting2 could not be found.", exc.message)

    @mock.patch(
        'nova.utils.get_sdk_adapter', side_effect=nova.utils.get_sdk_adapter)
    def test_get_access(self, mock_get_sdk_adapter):
        """Tests that we manage to get an access id based on access_type and
        access_to parameters.
        """
        access = self.api.get_access(self.context, '1234', 'ip', '0.0.0.0/0')
        mock_get_sdk_adapter.assert_called_once_with(
            "shared-file-system",
            admin=True,
            check_service=True,
            context=self.context,
            shared_file_system_api_version="2.82",
            global_request_id=self.context.global_id,
        )
        self.assertEqual('a25b2df3-90bd-4add-afa6-5f0dbbd50452', access.id)
        self.assertEqual('rw', access.access_level)
        self.assertEqual('active', access.state)
        self.assertEqual('ip', access.access_type)
        self.assertEqual('0.0.0.0/0', access.access_to)
        self.assertIsNone(access.access_key)

    def test_get_access_not_existing(self):
        """Tests that we get None if the access id does not exist.
        """
        access = self.api.get_access(
            self.context, "1234", "ip", "192.168.0.1/32"
        )

        self.assertIsNone(access)

    def test_allow_access_fails_non_existing_share(self):
        """Tests that we fail if trying to allow an
        non existing share.
        """
        exc = self.assertRaises(
            exception.ShareNotFound,
            self.api.allow,
            self.context,
            "nonexisting",
            "ip",
            "0.0.0.0/0",
            "rw",
        )

        self.assertIn("Share nonexisting could not be found.", exc.message)

    @mock.patch(
        'nova.utils.get_sdk_adapter', side_effect=nova.utils.get_sdk_adapter)
    def test_allow_access(self, mock_get_sdk_adapter):
        """Tests that we manage to allow access to a share.
        """
        access = self.api.allow(self.context, '1234', 'ip', '0.0.0.0/0', 'rw')
        mock_get_sdk_adapter.assert_called_once_with(
            "shared-file-system",
            admin=True,
            check_service=True,
            context=self.context,
            shared_file_system_api_version="2.82",
            global_request_id=self.context.global_id,
        )
        self.assertEqual('a25b2df3-90bd-4add-afa6-5f0dbbd50452', access.id)
        self.assertEqual('rw', access.access_level)
        self.assertEqual('active', access.state)
        self.assertEqual('ip', access.access_type)
        self.assertEqual('0.0.0.0/0', access.access_to)
        self.assertIsNone(access.access_key)

    def test_allow_access_fails_already_exists(self):
        """Tests that we have an exception is the share already exists.
        """
        exc = self.assertRaises(
            exception.ShareAccessGrantError,
            self.api.allow,
            self.context,
            '2345',
            'ip',
            '0.0.0.0/0',
            'rw'
        )

        self.assertIn(
            'Share access could not be granted to share',
            exc.message)

    def test_deny_access_fails_non_existing_share(self):
        """Tests that we fail if trying to deny an
        non existing share.
        """
        exc = self.assertRaises(
            exception.ShareNotFound,
            self.api.deny,
            self.context,
            "nonexisting",
            "ip",
            "0.0.0.0/0",
        )

        self.assertIn("Share nonexisting could not be found.", exc.message)

    @mock.patch(
        'nova.utils.get_sdk_adapter', side_effect=nova.utils.get_sdk_adapter)
    def test_deny_access(self, mock_get_sdk_adapter):
        """Tests that we manage to deny access to a share.
        """
        self.api.deny(
            self.context,
            '1234',
            'ip',
            '0.0.0.0/0'
        )
        self.assertEqual(2, mock_get_sdk_adapter.call_count)

        self.assertEqual(
            mock_get_sdk_adapter.call_args_list[0].args,
            (
                "shared-file-system",
            ),
        )
        self.assertEqual(
            mock_get_sdk_adapter.call_args_list[0].kwargs,
            {
                "admin": True,
                "check_service": True,
                "context": self.context,
                "shared_file_system_api_version": "2.82",
                "global_request_id": self.context.global_id,
            },
        )

        self.assertEqual(
            mock_get_sdk_adapter.call_args_list[1].args,
            (
                "shared-file-system",
            ),
        )
        self.assertEqual(
            mock_get_sdk_adapter.call_args_list[1].kwargs,
            {
                "admin": True,
                "check_service": True,
                "context": self.context,
                "shared_file_system_api_version": "2.82",
                "global_request_id": self.context.global_id,
            },
        )

    def test_deny_access_fails_id_missing(self):
        """Tests that we fail if something wrong happens calling deny method.
        """
        exc = self.assertRaises(exception.ShareAccessRemovalError,
                self.api.deny,
                self.context,
                '2345',
                'ip',
                '0.0.0.0/0'
                )

        self.assertIn(
            'Share access could not be removed from',
            exc.message)
        self.assertEqual(
            500,
            exc.code)

    def test_deny_access_fails_access_not_found(self):
        """Tests that we fail if access is missing.
        """
        exc = self.assertRaises(exception.ShareAccessNotFound,
                self.api.deny,
                self.context,
                '4567',
                'ip',
                '0.0.0.0/0'
                )

        self.assertIn(
            'Share access from Manila could not be found',
            exc.message)
        self.assertEqual(
            404,
            exc.code)
